"""Automatic differentation support for constructing derivative functions.

Multiple automatic differentiation backends are supported:

* `jax`: High performance array computing framework, with support for running
  computations on accelerator devices and just-in-time (JIT) compilation. To
  differentiate a function using the `jax` backend, the function must be defined in
  terms of JAX primitives, for example using the functions in the `jax.numpy` API. By
  default the derivative functions produced will be JIT compiled; a `jax_nojit` variant
  is available if the function (or its derivative) is not compatible with JIT
  compilation.
* `autograd`: Autograd can automatically differentiate native Python and NumPy code. To
  differentiate a function using the `autograd` backend it should be defined in terms
  of functions from the `autograd.numpy` and `autograd.scipy` APIs. Compared to JAX,
  the lack of JIT compilation in Autograd and features such as automatic vectorisation
  make Autograd slower, and so JAX will generally be a better choice.
* `symnum`: SymNum is a Python package that acts a bridge between NumPy and SymPy,
  providing a NumPy-like interface that can be used to symbolically define functions
  which take arrays as arguments and return arrays or scalars as values. To
  differentiate a function using the `symnum` backend it should be defined in terms of
  functions from the `symnum.numpy` API, and should have been decorated with
  ``symnum.numpify`` with specified argument shapes. SymNum is intended for use in
  generating the derivatives of 'simple' functions which compose a relatively small
  number of operations and act on small array inputs. By reducing interpreter overheads
  it can produce code which is cheaper to evaluate than corresponding Autograd or JAX
  functions (including those using JIT compilation) in such cases, and which can be
  serialised with the inbuilt Python pickle library allowing use for example in
  libraries which use multiprocessing to implement parallelisation across multiple
  processes.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, NamedTuple

from mici.autodiff import autograd_wrapper, jax_wrapper, symnum_wrapper

if TYPE_CHECKING:
    from collections.abc import Callable
    from types import ModuleType
    from typing import Optional


"""Names of valid differential operators.

Any automatic differentiation framework wrapper module will need to provide all of these
operators as callables (with a single function as argument) to fully support all of the
required derivative functions.
"""
DIFF_OPS = (
    # vector Jacobian product and value
    "vjp_and_value",
    # gradient and value for scalar valued functions
    "grad_and_value",
    # Hessian matrix, gradient and value for scalar valued functions
    "hessian_grad_and_value",
    # matrix Tressian product, gradient and value for scalar valued functions
    "mtp_hessian_grad_and_value",
    # Jacobian matrix and value for vector valued functions
    "jacobian_and_value",
    # matrix Hessian product, Jacobian matrix and value for vector valued functions
    "mhp_jacobian_and_value",
)


class AutodiffBackend(NamedTuple):
    """Automatic differentiation backend framework.

    Consists of a module defining differential operators, a boolean flag indicating if
    backend is available in current environment and optionally a function wrapper which
    applies any post processing required to functions.
    """

    module: ModuleType
    available: bool
    function_wrapper: Callable | None = None


"""Available autodifferentiation framework backends."""
_REGISTERED_BACKENDS = {
    "jax": AutodiffBackend(
        jax_wrapper,
        jax_wrapper.JAX_AVAILABLE,
        jax_wrapper.jit_and_return_numpy_arrays,
    ),
    "jax_nojit": AutodiffBackend(
        jax_wrapper,
        jax_wrapper.JAX_AVAILABLE,
        jax_wrapper.return_numpy_arrays,
    ),
    "autograd": AutodiffBackend(autograd_wrapper, autograd_wrapper.AUTOGRAD_AVAILABLE),
    "symnum": AutodiffBackend(symnum_wrapper, symnum_wrapper.SYMNUM_AVAILABLE),
}


def _get_backend(name: str) -> AutodiffBackend:
    # Normalize name string to all lowercase to make invariant to capitalization
    name = name.lower()
    if name not in _REGISTERED_BACKENDS:
        msg = (
            f"Selected autodiff backend {name} not recognised: "
            f"available options are {tuple(_REGISTERED_BACKENDS)}."
        )
        raise ValueError(msg)
    return _REGISTERED_BACKENDS[name]


def wrap_function(function: Callable, backend: str | None) -> Callable:
    """Apply function wrapper for automatic differentiation backend to a function.

    Backends may define a function wrapper which applies any post processing required to
    functions using framework - for example ensuring the function returns NumPy arrays
    or just-in-time compiling the function.

    Args:
        function: Function to wrap.
        backend: Name of automatic differentiation framework backend to use. If `None`
            function is returned unchanged.

    Returns:
        Wrapped function.
    """
    if backend is None:
        return function
    backend = _get_backend(backend)
    if backend.function_wrapper is not None:
        return backend.function_wrapper(function)
    return function


def autodiff_fallback(
    diff_func: Callable | None,
    func: Callable,
    diff_op_name: str,
    name: str,
    backend: str | None,
) -> Callable:
    """Generate derivative function automatically if not provided.

    Uses automatic differentiation to generate a function corresponding to a
    differential operator applied to a function if an alternative implementation of the
    derivative function has not been provided.

    Args:
        diff_func: Either a callable implementing the required derivative function or
            `None` if none was provided.
        func: Function to differentiate.
        diff_op_name: String specifying name of differential operator from automatic
            differentiation framework wrapper to use to generate required derivative
            function.
        name: Name of derivative function to use in error message.
        backend: Name of automatic differentiation framework backend to use. If `None`
            `diff_func` must be provided.

    Returns:
        `diff_func` value if not `None` otherwise generated derivative of `func` by
        applying named differential operator from automatic differentiation backend.
    """
    if diff_func is not None:
        return diff_func
    if diff_func is None and backend is None:
        msg = (
            f"Automatic differentiation backend specified as `None` so {name} must"
            "be provided directly."
        )
        raise ValueError(msg)
    if diff_op_name not in DIFF_OPS:
        msg = f"Differential operator {diff_op_name} is not defined."
        raise ValueError(msg)
    autodiff_backend = _get_backend(backend)
    if autodiff_backend.available:
        diff_func = getattr(autodiff_backend.module, diff_op_name)(func)
        return wrap_function(diff_func, backend)
    msg = (
        f"{backend} selected as autodiff backend but is not available in "
        f"current environment therefore {name} must be provided directly."
    )
    raise ValueError(msg)



"""MCMC samplers based on simulating Hamiltonian dynamics on a manifold."""

__authors__ = "Matt Graham"
__license__ = "MIT"

import mici.adapters
import mici.autodiff
import mici.integrators
import mici.interop
import mici.matrices
import mici.samplers
import mici.solvers
import mici.stagers
import mici.states
import mici.systems
import mici.transitions



"""Type aliases."""

from __future__ import annotations

from collections.abc import Callable, Collection, Iterable
from typing import Any, TypeAlias, TypeVar

from numpy import number
from numpy.typing import ArrayLike

from mici.matrices import Matrix, PositiveDefiniteMatrix
from mici.states import ChainState
from mici.systems import System
from mici.utils import LogRepFloat

ScalarLike: TypeAlias = bool | int | float | LogRepFloat | number
"""Scalar like objects."""

MatrixLike: TypeAlias = ArrayLike | Matrix
"""Matrix like objects."""

MetricLike: TypeAlias = ArrayLike | PositiveDefiniteMatrix
"""Metric (positive definite matrix) like objects."""

TransitionStatistics: TypeAlias = dict[str, ScalarLike]
"""Dictionary of statistics computed by :py:meth:`mici.transitions.Transition.sample`"""

AdaptationStatisticFunction: TypeAlias = Callable[[TransitionStatistics], float]
"""Function returning adaptation statistic given dictionary of transition statistics."""

ReducerFunction: TypeAlias = Callable[[Collection[float]], float]
"""Function combining per-chain control variables from adapters."""

AdapterState: TypeAlias = dict[str, Any]
"""Dictionary defining current state of an :py:class:`mici.adapters.Adapter`."""

ScalarFunction: TypeAlias = Callable[[ArrayLike], ScalarLike]
"""Function taking an array-like input and returning a scalar-like output."""

ArrayFunction: TypeAlias = Callable[[ArrayLike], ArrayLike]
"""Function taking an array-like input and returning a array-like output."""

GradientFunction: TypeAlias = Callable[
    [ArrayLike], ArrayLike | tuple[ArrayLike, ScalarLike]
]
"""Function returning the gradient of a scalar-valued function.

May optionally also return scalar-like value of  function.
"""

HessianFunction: TypeAlias = Callable[
    [ArrayLike],
    ArrayLike | tuple[ArrayLike, ArrayLike, ScalarLike],
]
"""Function returning the Hessian matrix of a scalar-valued function.

May optionally also return gradient and value of function.
"""

MatrixTressianProduct: TypeAlias = Callable[[ArrayLike], ArrayLike]
"""Function returning the product between a matrix and 'Tressian' of a scalar function.

Tressian here is the 3-dimensional array of third-order derivatives of a scalar
function.
"""

MatrixTressianProductFunction: TypeAlias = Callable[
    [ArrayLike],
    MatrixTressianProduct
    | tuple[MatrixTressianProduct, ArrayLike, ArrayLike, ScalarLike],
]
"""Function returning a matrix-Tressian product function for a scalar function.

Tressian here is the 3-dimensional array of third-order derivatives of a scalar
function.

May optionally also return the Hessian matrix, gradient and value of function.
"""

JacobianFunction: TypeAlias = Callable[
    [ArrayLike], ArrayLike | tuple[ArrayLike, ArrayLike]
]
"""Function returning the Jacobian of a array (vector) valued function.

May optionally aslo return the array value of the function.
"""

MatrixHessianProduct: TypeAlias = Callable[[ArrayLike], ArrayLike]
"""Function returning the product between a matrix and Hessian of an array function.

Hessian here is the 3-dimensional array of second-order derivatives of an array-valued
function.
"""

MatrixHessianProductFunction: TypeAlias = Callable[
    [ArrayLike],
    MatrixHessianProduct | tuple[MatrixHessianProduct, ArrayLike, ArrayLike],
]
"""Function returning a matrix-Hessian product function for an array function.

Hessian here is the 3-dimensional array of second-order derivatives of an array-valued
function.

May optionally also return the Jacobian matrix and array value of function.
"""

VectorJacobianProduct: TypeAlias = Callable[[ArrayLike], ArrayLike]
"""Function returning the product between an array and Jacobian of a array function.

Jacobian here is the (d+1)-dimensional array of first-order derivatives of an
d-dimensional array-valued function taking a 1-dimensional array as input.
"""

VectorJacobianProductFunction: TypeAlias = Callable[
    [ArrayLike],
    VectorJacobianProduct | tuple[VectorJacobianProduct, ArrayLike],
]
"""Function returning a vector-Jaccobian product function for an array function.

Jacobian here is the (d+1)-dimensional array of first-order derivatives of an
d-dimensional array-valued function taking a 1-dimensional array as input.

May optionally also return the array value of function.
"""

T = TypeVar("T")
PyTree = dict[Any, "PyTree"] | list["PyTree"] | tuple["PyTree"] | T
"""Arbitrarily nested structure of Python dict, list and tuple types."""

TraceFunction: TypeAlias = Callable[[ChainState], dict[str, ArrayLike]]
"""Function returning a dictionary of chain variables to trace given chain state."""

TerminationCriterion: TypeAlias = Callable[
    [System, ChainState, ChainState, ArrayLike], bool
]
"""Function indicating whether to terminate trajectory tree expansion."""

ChainIterator: TypeAlias = Iterable[tuple[int, dict]]

SystemStateMethod: TypeAlias = Callable[[System, "ChainState"], ArrayLike | ScalarLike]
"""System method computing function of chain state."""

SystemStateWithAuxMethod: TypeAlias = Callable[
    [System, ChainState], tuple[ArrayLike | ScalarLike, ...]
]
"""System method computing function chain state with auxiliary return values."""



"""Symplectic integrators for simulation of Hamiltonian dynamics."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

import numpy as np

from mici.errors import AdaptationError, NonReversibleStepError
from mici.solvers import (
    FixedPointSolver,
    ProjectionSolver,
    maximum_norm,
    solve_fixed_point_direct,
    solve_projection_onto_manifold_newton,
)

if TYPE_CHECKING:
    from collections.abc import Callable, Sequence
    from typing import Any

    from numpy.typing import ArrayLike

    from mici.states import ChainState
    from mici.systems import ConstrainedTractableFlowSystem, System, TractableFlowSystem
    from mici.types import NormFunction


class Integrator(ABC):
    r"""Base class for integrators for simulating Hamiltonian dynamics.

    For a Hamiltonian function :math:`h` with position variables :math:`q` and momentum
    variables :math:`p`, the canonical Hamiltonian dynamic is defined by the ordinary
    differential equation system

    .. math::

        \dot{q} = \nabla_2 h(q, p),  \qquad \dot{p} = -\nabla_1 h(q, p),

    with the flow map :math:`\Phi` corresponding to the solution of the corresponding
    initial value problem a time-reversible and symplectic (and by consequence
    volume-preserving) map.

    Derived classes implement a :py:meth:`step` method which approximates the flow-map
    with :math:`\Psi(t) \approx \Phi(t)` over some small time interval :math:`t`, while
    conserving the properties of being time-reversible and symplectic, with composition
    of this integrator step method allowing simulation of time-discretised trajectories
    of the Hamiltonian dynamics.
    """

    def __init__(self, system: System, step_size: float | None = None) -> None:
        """
        Args:
            system: Hamiltonian system to integrate the dynamics of.
            step_size: Integrator time step. If set to :code:`None` it is assumed that a
                step size adapter will be used to set the step size before calling the
                :py:meth:`step` method.
        """
        self.system = system
        self.step_size = step_size

    def step(self, state: ChainState) -> ChainState:
        """Perform a single integrator step from a supplied state.

        Args:
            state: System state to perform integrator step from.

        Returns:
            New object corresponding to stepped state.
        """
        if self.step_size is None:
            msg = (
                "Integrator `step_size` is `None`. This value should only be used if a "
                "step size adapter is being used to set the step size."
            )
            raise AdaptationError(msg)
        state = state.copy()
        self._step(state, state.dir * self.step_size)
        return state

    @abstractmethod
    def _step(self, state: ChainState, time_step: float) -> None:
        """Implementation of single integrator step.

        Args:
            state: System state to perform integrator step from. Updated in place.
            time_step: Integrator time step. May be positive or negative.
        """


class TractableFlowIntegrator(Integrator):
    r"""Base class for integrators for Hamiltonian systems with tractable flows.

    The Hamiltonian function is assumed to be expressible as the sum of two analytically
    tractable components for which the corresponding Hamiltonian flows can be exactly
    simulated. Specifically it is assumed that the Hamiltonian function :math:`h` takes
    the form

    .. math::

        h(q, p) = h_1(q) + h_2(q, p),

    where :math:`q` and :math:`p` are the position and momentum variables respectively,
    and :math:`h_1` and :math:`h_2` are Hamiltonian component functions for which the
    exact flow maps, :math:`\Phi_1` and :math:`\Phi_2` respectively, can be computed
    exactly.
    """

    def __init__(
        self, system: TractableFlowSystem, step_size: float | None = None
    ) -> None:
        """
        Args:
            system: Hamiltonian system to integrate the dynamics of with tractable
                Hamiltonian component flows.
            step_size: Integrator time step. If set to :code:`None` it is assumed that a
                step size adapter will be used to set the step size before calling the
                :py:meth:`step` method.
        """
        if not hasattr(system, "h1_flow") or not hasattr(system, "h2_flow"):
            msg = (
                f"{type(self)} can only be used for systems with explicit `h1_flow` "
                f"and `h2_flow` Hamiltonian component flow maps. For systems in which "
                f"only `h1_flow` is available the `ImplicitLeapfrogIntegrator` class "
                f"may be used instead and for systems in which neither `h1_flow` or "
                f"`h2_flow` is available the `ImplicitMidpointIntegrator` class may be "
                f"used."
            )
            raise ValueError(msg)
        super().__init__(system, step_size)


class LeapfrogIntegrator(TractableFlowIntegrator):
    r"""Leapfrog integrator for Hamiltonian systems with tractable component flows.

    The overall integrator step :math:`\Psi` is defined by the symmetric composition

    .. math::

        \Psi(t) = \Phi_1(t/2) \circ \Phi_2(t) \circ \Phi_1(t/2)

    where :math:`\Phi_1` and :math:`\Phi_2` are the exact flow maps associated with the
    Hamiltonian components :math:`h_1` and :math:`h_2` respectively.

    For separable Hamiltonians of the

    .. math::

        h(q, p) = h_1(q) + h_2(p),

    where :math:`h_1` is the potential energy and :math:`h_2` is the kinetic energy,
    this integrator corresponds to the classic (position) Störmer-Verlet method.

    The integrator can also be applied to the more general Hamiltonian splitting

    .. math::

        h(q, p) = h_1(q) + h_2(q, p),

    providing the flows for :math:`h_1` and :math:`h_2` are both tractable.

    For more details see Sections 2.6 and 4.2.2 in Leimkuhler and Reich (2004).

    References:
      1. Leimkuhler, B., & Reich, S. (2004). Simulating Hamiltonian Dynamics (No. 14).
         Cambridge University Press.
    """

    def _step(self, state: ChainState, time_step: float) -> None:
        self.system.h1_flow(state, 0.5 * time_step)
        self.system.h2_flow(state, time_step)
        self.system.h1_flow(state, 0.5 * time_step)


class SymmetricCompositionIntegrator(TractableFlowIntegrator):
    r"""Symmetric composition integrator for Hamiltonians with tractable flows.

    The Hamiltonian function is assumed to be expressible as the sum of two analytically
    tractable components for which the corresponding Hamiltonian flows can be exactly
    simulated. Specifically it is assumed that the Hamiltonian function :math:`h` takes
    the form

    .. math::

        h(q, p) = h_1(q) + h_2(q, p),

    where :math:`q` and :math:`p` are the position and momentum variables respectively,
    and :math:`h_1` and :math:`h_2` are Hamiltonian component functions for which the
    exact flows, respectively :math:`\Phi_1` and :math:`\Phi_2`, can be computed. An
    alternating composition can then be formed as

    .. math::

        \Psi(t) = A(a_S t) \circ B(b_S t) \circ \dots \circ
                  A(a_1 t) \circ B(b_1 t) \circ A(a_0 t),

    where :math:`A = \Phi_1` and :math:`B = \Phi_2` or :math:`A = \Phi_2` and :math:`B =
    \Phi_1`, and :math:`(a_0,\dots, a_S)` and :math:`(b_1, \dots, b_S)` are a set of
    coefficients to be determined with :math:`S \geq 1`.

    To ensure a consistency (i.e. the integrator is at least order one) we require that

    .. math::

        \sum_{s=0}^S a_s = \sum_{s=1}^S b_s = 1.

    For symmetric compositions we restrict that

    .. math::

        a_{S-m} = a_m, \quad b_{S+1-m} = b_m,

    with symmetric consistent methods of at least order two.

    The combination of the symmetry and consistency requirements mean that for each
    :math:`S \geq 1` a symmetric composition method can be described by :math:`S - 1`
    'free' coefficients :math:`(a_0, b_1, \dots, a_{K-1}, b_K)`  with
    :math:`K = (S - 1) / 2` if :math:`S > 1` is odd (with no free coefficients for
    :math:`S = 1` case) or :math:`(a_0, b_1, \dots, a_K)` with :math:`K = (S - 2) / 2`
    if :math:`S > 2` is even (with a single free coefficient :math:`a_0` for :math:`S=2`
    case).

    The Störmer-Verlet 'leapfrog' integrator is the special case corresponding to the
    unique (symmetric and consistent) '1-stage' (:math:`S = 1`) integrator.

    For more details see Section 6.2 in Leimkuhler and Reich (2004).

    References:
      1. Leimkuhler, B., & Reich, S. (2004). Simulating Hamiltonian Dynamics (No. 14).
         Cambridge University Press.
    """

    def __init__(
        self,
        system: TractableFlowSystem,
        free_coefficients: Sequence[float],
        *,
        step_size: float | None = None,
        initial_h1_flow_step: bool = True,
    ) -> None:
        r"""
        Args:
            system: Hamiltonian system to integrate the dynamics of with tractable
                Hamiltonian component flows.
            free_coefficients: Sequence of :math:`S - 1` scalar values, where :math:`S`
                is the number of stages in the symmetric composition, specifying the
                free coefficients :math:`(a_0, b_1, a_1, \dots, a_K, b_K)` with :math:`K
                = (S - 1) / 2` if :math:`S` is odd or :math:`(a_0, b_1, a_1, \dots,
                a_K)` with :math:`k = (S - 2) / 2` if `S` is even.
            step_size: Integrator time step. If set to :code:`None` it is assumed that a
                step size adapter will be used to set the step size before calling the
                :py:meth:`step` method.
            initial_h1_flow_step: Whether the initial :math:`A` flow in the composition
                should correspond to the flow of the `h_1` Hamiltonian component
                (:code:`True`) or to the flow of the :math:`h_2` component
                (:code:`False`).
        """
        super().__init__(system, step_size)
        self.initial_h1_flow_step = initial_h1_flow_step
        n_free_coefficients = len(free_coefficients)
        coefficients = list(free_coefficients)
        coefficients.append(
            0.5 - sum(free_coefficients[(n_free_coefficients) % 2 :: 2]),
        )
        coefficients.append(
            1 - 2 * sum(free_coefficients[(n_free_coefficients + 1) % 2 :: 2]),
        )
        self.coefficients = coefficients + coefficients[-2::-1]
        flow_a = system.h1_flow if initial_h1_flow_step else system.h2_flow
        flow_b = system.h2_flow if initial_h1_flow_step else system.h1_flow
        self.flows = [flow_a, flow_b] * (n_free_coefficients + 1) + [flow_a]

    def _step(self, state: ChainState, time_step: float) -> None:
        for coefficient, flow in zip(self.coefficients, self.flows, strict=True):
            flow(state, coefficient * time_step)


class BCSSTwoStageIntegrator(SymmetricCompositionIntegrator):
    r"""Two-stage symmetric composition integrator due to Blanes, Casas & Sanz-Serna.

    Described in equation (6.4) in Blanes, Casas, Sanz-Serna (2014).

    Corresponds to specific instance of :py:class:`SymmetricCompositionIntegrator` with
    :math:`S = 2` and free coefficient :math:`a_0 = (3 - \sqrt{3}) / 6`.

    References:
      1. Blanes, S., Casas, F., & Sanz-Serna, J. M. (2014).
         Numerical integrators for the Hybrid Monte Carlo method.
         SIAM Journal on Scientific Computing, 36(4), A1556-A1580.
    """

    def __init__(
        self, system: TractableFlowSystem, step_size: float | None = None
    ) -> None:
        """
        Args:
            system: Hamiltonian system to integrate the dynamics of with tractable
                Hamiltonian component flows.
            step_size: Integrator time step. If set to :code:`None` it is assumed that a
                step size adapter will be used to set the step size before calling the
                :py:meth:`step` method.
        """
        a_0 = (3 - 3**0.5) / 6
        super().__init__(system, (a_0,), step_size=step_size, initial_h1_flow_step=True)


class BCSSThreeStageIntegrator(SymmetricCompositionIntegrator):
    """Three-stage symmetric composition integrator due to Blanes, Casas & Sanz-Serna.

    Described in equation (6.7) in Blanes, Casas, Sanz-Serna (2014).

    Corresponds to specific instance of :py:class:`SymmetricCompositionIntegrator` with
    :math:`S = 3` and free coefficients :math:`a_0 = 0.11888010966548` and
    :math:`b_1 = 0.29619504261126`.

    References:
      1. Blanes, S., Casas, F., & Sanz-Serna, J. M. (2014).
         Numerical integrators for the Hybrid Monte Carlo method.
         SIAM Journal on Scientific Computing, 36(4), A1556-A1580.
    """

    def __init__(
        self, system: TractableFlowSystem, step_size: float | None = None
    ) -> None:
        """
        Args:
            system: Hamiltonian system to integrate the dynamics of with tractable
                Hamiltonian component flows.
            step_size: Integrator time step. If set to :code:`None` it is assumed that a
                step size adapter will be used to set the step size before calling the
                :py:meth:`step` method.
        """
        a_0 = 0.11888010966548
        b_1 = 0.29619504261126
        super().__init__(
            system,
            (a_0, b_1),
            step_size=step_size,
            initial_h1_flow_step=True,
        )


class BCSSFourStageIntegrator(SymmetricCompositionIntegrator):
    """Four-stage symmetric composition integrator due to Blanes, Casas & Sanz-Serna.

    Described in equation (6.8) in Blanes, Casas, Sanz-Serna (2014).

    Corresponds to specific instance of :py:class:`SymmetricCompositionIntegrator` with
    :math:`S = 4` and free coefficients :math:`a_0 = 0.071353913450279725904`,
    :math:`b_1 = 0.191667800000000000000` and :math:`a_1 = 0.268548791161230105820`.

    References:
      1. Blanes, S., Casas, F., & Sanz-Serna, J. M. (2014).
         Numerical integrators for the Hybrid Monte Carlo method.
         SIAM Journal on Scientific Computing, 36(4), A1556-A1580.
    """

    def __init__(
        self, system: TractableFlowSystem, step_size: float | None = None
    ) -> None:
        """
        Args:
            system: Hamiltonian system to integrate the dynamics of with tractable
                Hamiltonian component flows.
            step_size: Integrator time step. If set to :code:`None` it is assumed that
                a step size adapter will be used to set the step size before calling the
                :py:meth:`step` method.
        """
        a_0 = 0.071353913450279725904
        b_1 = 0.191667800000000000000
        a_1 = 0.268548791161230105820
        super().__init__(
            system,
            (a_0, b_1, a_1),
            step_size=step_size,
            initial_h1_flow_step=True,
        )


class ImplicitLeapfrogIntegrator(Integrator):
    r"""Implicit leapfrog integrator for Hamiltonians with a non-separable component.

    Also known as the generalised leapfrog method.

    The Hamiltonian function :math:`h` is assumed to take the form

    .. math::

        h(q, p) = h_1(q) + h_2(q, p),

    where :math:`q` and :math:`p` are the position and momentum variables respectively,
    :math:`h_1` is a Hamiltonian component function for which the exact flow can be
    computed and :math:`h_2` is a Hamiltonian component function of the position and
    momentum variables, which may be non-separable and for which exact simulation of the
    correspond Hamiltonian flow may not be possible.

    The overall integrator step :math:`\Psi` is defined by the symmetric composition

    .. math::

        \Psi(t) =
        A(t/2) \circ B(t/2) \circ C(t/2) \circ C^*(t/2) \circ B^*(t/2) \circ A^*(t/2),

    where the *adjoint* of a flow map :math:`X` is defined such that :math:`X^*(t) =
    X(-t)^{-1}` and the component maps are defined by

    .. math::

        A(t)(q, p) = A^*(t)(q, p) = (q, p - t\nabla h_1(q)), \\
        B(t)(q, p) = \lbrace (q, p') : p' = p - t \nabla_1 h_2(q, p') \rbrace, \\
        B^*(t)(q, p) = (q, p - t \nabla_1 h_2(q, p)), \\
        C(t)(q, p) = (q + t \nabla_2 h_2(q, p), p), \\
        C^*(t)(q, p) = \lbrace (q', p) : q' = q + t \nabla_2 h_2(q', p) \rbrace.

    The resulting implicit integrator is a symmetric second-order method corresponding
    to a symplectic partitioned Runge-Kutta method. See Section 6.3.2 in Leimkuhler and
    Reich (2004) for more details.

    Fixed-point iterations are used to solve the non-linear systems of equations in the
    implicit component updates (:math:`B` and :math:`C^*`). As the iterative solves may
    fail to converge, or may converge to one of multiple solutions, following the
    approach proposed by Zappa, Holmes-Cerfon and Goodman (2018), an explicit
    *reversibility check* is performed to ensure the overall integrator step is
    time-reversible. If the reversibility check fails or the iterative solver fails to
    converge an appropriate error is raised
    (:py:exc:`mici.errors.NonReversibleStepError` and
    :py:exc:`mici.errors.ConvergenceError` respectively).

    References:
      1. Leimkuhler, B., & Reich, S. (2004). Simulating Hamiltonian Dynamics (No. 14).
         Cambridge University Press.
      2. Zappa, E., Holmes-Cerfon, M., & Goodman, J. (2018). Monte Carlo on manifolds:
         sampling densities and integrating functions. Communications on Pure and
         Applied Mathematics, 71(12), 2609-2647.
    """

    def __init__(
        self,
        system: System,
        step_size: float | None = None,
        reverse_check_tol: float = 2e-8,
        reverse_check_norm: NormFunction = maximum_norm,
        fixed_point_solver: FixedPointSolver = solve_fixed_point_direct,
        fixed_point_solver_kwargs: dict[str, Any] | None = None,
    ) -> None:
        """
        Args:
            system: Hamiltonian system to integrate the dynamics of.
            step_size: Integrator time step. If set to :code:`None` it is assumed that a
                step size adapter will be used to set the step size before calling the
                :py:meth:`step` method.
            reverse_check_tol: Tolerance for check of reversibility of implicit
                sub-steps which involve iterative solving of a non-linear system of
                equations. The step is assumed to be reversible if sequentially applying
                the forward and adjoint updates to a state returns to a state with a
                position component within a distance (defined by the
                :code:`reverse_check_norm` argument) of :code:`reverse_check_tol` of the
                original state position component. If this condition is not met a
                :py:exc:`mici.errors.NonReversibleStepError` exception is raised.
            reverse_check_norm: Norm function accepting a single one-dimensional array
                input and returning a non-negative floating point value defining the
                distance to use in the reversibility check. Defaults to
                :py:func:`mici.solvers.maximum_norm`.
            fixed_point_solver:  Function which given a function :code:`func` and
                initial guess :code:`x0` iteratively solves the fixed point equation
                :code:`func(x) = x` initialising the iteration with :code:`x0` and
                returning an array corresponding to the solution if the iteration
                converges or raising a :py:class:`mici.errors.ConvergenceError`
                otherwise. Defaults to :py:exc:`mici.solvers.solve_fixed_point_direct`.
            fixed_point_solver_kwargs: Dictionary of any keyword arguments to
                :code:`fixed_point_solver`.
        """
        super().__init__(system, step_size)
        self.reverse_check_tol = reverse_check_tol
        self.reverse_check_norm = reverse_check_norm
        self.fixed_point_solver = fixed_point_solver
        if fixed_point_solver_kwargs is None:
            fixed_point_solver_kwargs = {}
        self.fixed_point_solver_kwargs = fixed_point_solver_kwargs

    def _solve_fixed_point(
        self,
        fixed_point_func: Callable[[ArrayLike], ArrayLike],
        x_init: ArrayLike,
    ) -> ArrayLike:
        return self.fixed_point_solver(
            fixed_point_func,
            x_init,
            **self.fixed_point_solver_kwargs,
        )

    def _step_a(self, state: ChainState, time_step: float) -> None:
        self.system.h1_flow(state, time_step)

    def _step_b_fwd(self, state: ChainState, time_step: float) -> None:
        def fixed_point_func(mom: ArrayLike) -> ArrayLike:
            state.mom = mom
            return mom_init - time_step * self.system.dh2_dpos(state)

        mom_init = state.mom
        state.mom = self._solve_fixed_point(fixed_point_func, mom_init)

    def _step_b_adj(self, state: ChainState, time_step: float) -> None:
        mom_init = state.mom.copy()
        state.mom -= time_step * self.system.dh2_dpos(state)
        state_back = state.copy()
        self._step_b_fwd(state_back, -time_step)
        rev_diff = self.reverse_check_norm(state_back.mom - mom_init)
        if rev_diff > self.reverse_check_tol:
            msg = (
                f"Non-reversible step. Distance between initial and forward-backward "
                f"integrated momentums = {rev_diff:.1e}."
            )
            raise NonReversibleStepError(msg)

    def _step_c_fwd(self, state: ChainState, time_step: float) -> None:
        pos_init = state.pos.copy()
        state.pos += time_step * self.system.dh2_dmom(state)
        state_back = state.copy()
        self._step_c_adj(state_back, -time_step)
        rev_diff = self.reverse_check_norm(state_back.pos - pos_init)
        if rev_diff > self.reverse_check_tol:
            msg = (
                f"Non-reversible step. Distance between initial and forward-backward "
                f"integrated positions = {rev_diff:.1e}."
            )
            raise NonReversibleStepError(msg)

    def _step_c_adj(self, state: ChainState, time_step: float) -> None:
        def fixed_point_func(pos: ArrayLike) -> ArrayLike:
            state.pos = pos
            return pos_init + time_step * self.system.dh2_dmom(state)

        pos_init = state.pos
        state.pos = self._solve_fixed_point(fixed_point_func, pos_init)

    def _step(self, state: ChainState, time_step: float) -> None:
        self._step_a(state, time_step)
        self._step_b_fwd(state, time_step)
        self._step_c_fwd(state, time_step)
        self._step_c_adj(state, time_step)
        self._step_b_adj(state, time_step)
        self._step_a(state, time_step)


class ImplicitMidpointIntegrator(Integrator):
    r"""Implicit midpoint integrator for general Hamiltonians.

    The Hamiltonian function :math:`h` may be a general (non-separable) function of both
    the position variables :math:`q` and momentum variables :math:`p`.

    The Hamiltonian flow :math:`\Phi` is approximated with the symmetric composition
    :math:`\Psi(t) = A(t/2) \circ A^*(t/2)` of an implicit Euler half-step
    :math:`A(t/2)` with an explicit Euler half-step :math:`A^*(t/2)` (which is adjoint
    to the implicit Euler step, that is :math:`A^*(t) = A(-t)^{-1}`), with the
    components maps defined by

    .. math::

        A(t)(q, p) = \lbrace
        (q', p') :
        q' = q +  t \nabla_2 h(q', p'),
        p' = p - t \nabla_1 h(q', p'))
        \rbrace, \\
        A^*(t)(q, p) = (q +  t \nabla_2 h(q, p), p - t \nabla_1 h(q, p)).

    The resulting implicit integrator is a second-order method corresponding to a
    symplectic one-stage Runge-Kutta method. See Sections 4.1 and 6.3.1 in Leimkuhler
    and Reich (2004) for more details.

    A fixed-point iteration is used to solve the non-linear system of equations in the
    implicit Euler step :math:`A`. As the iterative solve may fail to converge, or may
    converge to one of multiple solutions, following the approach proposed by
    Zappa, Holmes-Cerfon and Goodman (2018), an explicit *reversibility check* is
    performed to ensure the overall integrator step is time-reversible. If the
    reversibility check fails or the iterative solver fails to converge an appropriate
    error is raised (:py:exc:`mici.errors.NonReversibleStepError` and
    :py:exc:`mici.errors.ConvergenceError` respectively).

    References:
      1. Leimkuhler, B., & Reich, S. (2004). Simulating Hamiltonian Dynamics (No. 14).
         Cambridge University Press.
      2. Zappa, E., Holmes-Cerfon, M., & Goodman, J. (2018). Monte Carlo on manifolds:
         sampling densities and integrating functions. Communications on Pure and
         Applied Mathematics, 71(12), 2609-2647.
    """

    def __init__(
        self,
        system: System,
        step_size: float | None = None,
        reverse_check_tol: float = 2e-8,
        reverse_check_norm: NormFunction = maximum_norm,
        fixed_point_solver: FixedPointSolver = solve_fixed_point_direct,
        fixed_point_solver_kwargs: dict[str, Any] | None = None,
    ) -> None:
        """
        Args:
            system: Hamiltonian system to integrate the dynamics of.
            step_size: Integrator time step. If set to :code:`None` it is assumed that a
                step size adapter will be used to set the step size before calling the
                `step` method.
            reverse_check_tol: Tolerance for check of reversibility of implicit Euler
                steps which involve iterative solving of a non-linear system of
                equations. The step is assumed to be reversible if sequentially applying
                the forward and adjoint updates to a state returns to a state with a
                position component within a distance (defined by the
                :code:`reverse_check_norm` argument) of :code:`reverse_check_tol` of the
                original state position component. If this condition is not met a
                :py:exc:`mici.errors.NonReversibleStepError` exception is raised.
            reverse_check_norm: Norm function accepting a single one-dimensional array
                input and returning a non-negative floating point value defining the
                distance to use in the reversibility check. Defaults to
                :py:func:`mici.solvers.maximum_norm`.
            fixed_point_solver: Function which given a function :code:`func` and initial
                guess :code:`x0` iteratively solves the fixed point equation
                :code:`func(x) = x` initialising the iteration with :code:`x0` and
                returning an array corresponding to the solution if the iteration
                converges or raising a :py:exc:`mici.errors.ConvergenceError`
                otherwise. Defaults to :py:func:`mici.solvers.solve_fixed_point_direct`.
            fixed_point_solver_kwargs: Dictionary of any keyword arguments to
                :code:`fixed_point_solver`.
        """
        super().__init__(system, step_size)
        self.reverse_check_tol = reverse_check_tol
        self.reverse_check_norm = reverse_check_norm
        self.fixed_point_solver = fixed_point_solver
        if fixed_point_solver_kwargs is None:
            fixed_point_solver_kwargs = {}
        self.fixed_point_solver_kwargs = fixed_point_solver_kwargs

    def _solve_fixed_point(
        self,
        fixed_point_func: Callable[[ArrayLike], ArrayLike],
        x_init: ArrayLike,
    ) -> ArrayLike:
        return self.fixed_point_solver(
            fixed_point_func,
            x_init,
            **self.fixed_point_solver_kwargs,
        )

    def _step_a_fwd(self, state: ChainState, time_step: float) -> None:
        pos_mom_init = np.concatenate([state.pos, state.mom])

        def fixed_point_func(pos_mom: ArrayLike) -> ArrayLike:
            state.pos, state.mom = np.split(pos_mom, 2)
            return pos_mom_init + np.concatenate(
                [
                    time_step * self.system.dh_dmom(state),
                    -time_step * self.system.dh_dpos(state),
                ],
            )

        state.pos, state.mom = np.split(
            self._solve_fixed_point(fixed_point_func, pos_mom_init),
            2,
        )

    def _step_a_adj(self, state: ChainState, time_step: float) -> None:
        state_prev = state.copy()
        state.pos += time_step * self.system.dh_dmom(state_prev)
        state.mom -= time_step * self.system.dh_dpos(state_prev)
        state_back = state.copy()
        self._step_a_fwd(state_back, -time_step)
        rev_diff = self.reverse_check_norm(
            np.concatenate(
                [state_back.pos - state_prev.pos, state_back.mom - state_prev.mom],
            ),
        )
        if rev_diff > self.reverse_check_tol:
            msg = (
                f"Non-reversible step. Distance between initial and forward-backward "
                f"integrated (pos, mom) pairs = {rev_diff:.1e}."
            )
            raise NonReversibleStepError(msg)

    def _step(self, state: ChainState, time_step: float) -> None:
        self._step_a_fwd(state, time_step / 2)
        self._step_a_adj(state, time_step / 2)


class ConstrainedLeapfrogIntegrator(TractableFlowIntegrator):
    r"""Leapfrog integrator for constrained Hamiltonian systems.

    The Hamiltonian function is assumed to be expressible as the sum of two components
    for which the corresponding (unconstrained) Hamiltonian flows can be exactly
    simulated. Specifically it is assumed that the Hamiltonian function `h` takes the
    form

    .. math::

        h(q, p) = h_1(q) + h_2(q, p),

    where :math:`q` and :math:`p` are the position and momentum variables respectively,
    and :math:`h_1` and :math:`h_2` Hamiltonian component functions for which the exact
    flows, respectively :math:`\Phi_1` and :math:`\Phi_2`, can be computed.

    The system is assumed to be additionally subject to a set of holonomic constraints
    on the position component of the state i.e. that all valid states must satisfy

    .. math::

        c(q) = 0,

    for some differentiable and surjective vector constraint function :math:`c` and the
    set of positions satisfying the constraints implicitly defining a manifold the
    dynamics remain confined to.

    The constraints are enforced by introducing a set of Lagrange multipliers
    :math:`\lambda` of dimension equal to number of constraints, and defining a
    'constrained' Hamiltonian

    .. math::

        \bar{h}(q, p) = h_1(q) + h_2(q, p) + c(q)^T\lambda  ~\text{s.t.}~ c(q) = 0,

    with corresponding dynamics described by the system of differential algebraic
    equations

    .. math::

        \dot{q} = \nabla_2 h(q, p), \quad
        \dot{p} = -\nabla_1 h(q, p) - \partial c(q)^T\lambda, \quad
        c(q) = 0.

    The dynamics implicitly define a set of constraints on the momentum variables,
    differentiating the constraint equation with respect to time giving that

    .. math::

        \partial c(q) \nabla_2 h(q, p) = \partial c(q) \nabla_2 h_2(q, p) = 0.

    The set of momentum variables satisfying the above for given position variables is
    termed the cotangent space of the manifold (at a position), and the set of
    position-momentum pairs for which the position is on the constraint manifold and the
    momentum in the corresponding cotangent space is termed the cotangent bundle.

    To define a second-order symmetric integrator which exactly (up to floating point
    error) preserves these constraints, forming a symplectic map on the cotangent
    bundle, we follow the approach of Reich (1996).

    We first define a map :math:`\Pi` parametrised by a vector of Lagrange multipliers
    :math:`\lambda`

    .. math::

        \Pi(\lambda)(q, p) = (q, p + \partial c(q)^T \lambda),

    with :math:`\lambda` allowed to be an implicitly defined function of :math:`q` and
    :math:`p`.

    We then define a map :math:`A` in terms of the :math:`h_1` flow map :math:`\Phi_1`
    as

    .. math::

        A(t) = \Pi(\lambda) \circ \Phi_1(t),

    with :math:`\lambda` implicitly defined such that for :math:`(q', p') = A(t)(q, p)`
    we have that :math:`\partial c(q') \nabla_2 h_2(q', p') = 0` for any initial state
    :math:`(q, p)` in the co-tangent bundle, with :math:`c(q') = 0` trivially satisfied
    as :math:`\Phi_1` is an identity map in the position:

    .. math::

        \Phi_1(t)(q, p) = (q, p - t \nabla h_1(q)).

    The map :math:`A(t)` therefore corresponds to taking an unconstrained step according
    to the :math:`h_1` component flow map :math:`\Phi_1(t)` and then projecting the
    resulting updated momentum back in to the co-tangent space. For the usual case in
    which :math:`h` includes only quadratic terms in the momentum :math:`p` such that
    :math:`\nabla_2 h(q, p)` is a linear function of :math:`p`, then :math:`\lambda`
    can be analytically solved for to give a closed-form expression for the projection
    into the co-tangent space.

    We also define a map :math:`B` in terms of the :math:`h_2` flow map :math:`\Phi_2`
    as

    .. math::

        B(t) = \Pi(\lambda') \circ \Phi_2(t) \circ \Pi(\lambda),

    such that for :math:`(q', p') = B(t)(q, p)`, :math:`\lambda` is implicitly defined
    such that :math:`c(q') = 0` and :math:`\lambda'` is implicitly defined such that
    :math:`\partial c(q') \nabla_2 h(q', p') = 0`.

    This can be decomposed as first solving for :math:`\lambda` such that

    .. math::

        c((\Phi_2(t) \circ \Pi(\lambda)(q, p))_1)
        = c((\Phi_2(t)(q, p + \partial c(q)^T \lambda))_1) = 0,

    i.e. solving for the values of the Lagrange multipliers such that the position
    component of the output of :math:`\Phi_2(t) \circ \Pi(\lambda)` is on the manifold,
    with this typically a non-linear system of equations that will need to be solved
    iteratively e.g. using Newton's method. The momentum output of :math:`\Phi_2(t)
    \circ \Pi(\lambda)` is then projected in to the cotangent space to compute the final
    state pair, with this projection step as noted above typically having an analytic
    solution.

    For more details see Reich (1996) and section 7.5.1 in Leimkuhler and Reich (2004).

    The overall second-order integrator is then defined as the symmetric composition

    .. math::

        \Psi(t) = A(t / 2) \circ B(t / N)^N \circ A(t / 2),

    where :math:`N` is a positive integer corresponding to the number of 'inner'
    :math:`h_2` flow steps, following the 'geodesic integrator' formulation proposed by
    Leimkuhler and Matthews (2016). The additional flexibility introduced by having
    the possibility of :math:`N > 1` is particularly of use when evaluation of
    :math:`\Phi_1` is significantly more expensive than evaluation of :math:`\Phi_2`; in
    this case using :math:`N > 1` can allow a larger time step :math:`t` to be used than
    may be otherwise possible due to the need to ensure the iterative solver used in
    :math:`B` does not diverge, with a smaller step size :math:`t / N` used for the
    steps involving the iterative solves with the (cheaper) :math:`\Phi_2` flow map
    and a larger step size :math:`t` used for the steps involving the (more expensive)
    :math:`\Phi_1` flow map.

    This integrator exactly preserves the constraints at all steps, such that if an
    initial position momentum pair :math:`(q, p)` are in the cotangent bundle, the
    corresponding pair after calling the :py:meth:`step` method of the integrator will
    also be in the cotangent bundle, *providing the iterative solver converges*.

    As the iterative solves may fail to converge, or may converge to one of multiple
    solutions, following the approach proposed by Zappa, Holmes-Cerfon and Goodman
    (2018), an explicit *reversibility check* is performed to ensure the overall
    integrator step is time-reversible; see also Lelievre, Rousset and Stoltz (2019) for
    an analysis of this approach specifically in the context of Hamiltonian Monte Carlo.
    If the reversibility check fails or the iterative solver fails to converge an
    appropriate error is raised (:py:exc:`mici.errors.NonReversibleStepError` and
    :py:exc:`mici.errors.ConvergenceError` respectively).

    References:
      1. Reich, S. (1996). Symplectic integration of constrained Hamiltonian systems by
         composition methods. SIAM journal on numerical analysis, 33(2), 475-491.
      2. Leimkuhler, B., & Reich, S. (2004). Simulating Hamiltonian Dynamics (No. 14).
         Cambridge University Press.
      3. Leimkuhler, B., & Matthews, C. (2016). Efficient molecular dynamics using
         geodesic integration and solvent-solute splitting. Proceedings of the Royal
         Society A: Mathematical, Physical and Engineering Sciences, 472(2189),
         20160138.
      4. Zappa, E., Holmes-Cerfon, M., & Goodman, J. (2018). Monte Carlo on manifolds:
         sampling densities and integrating functions. Communications on Pure and
         Applied Mathematics, 71(12), 2609-2647.
      5. Lelievre, T., Rousset, M., & Stoltz, G. (2019). Hybrid Monte Carlo methods for
         sampling probability measures on submanifolds. Numerische Mathematik, 143,
         379-421.
    """

    def __init__(
        self,
        system: ConstrainedTractableFlowSystem,
        step_size: float | None = None,
        n_inner_step: int = 1,
        reverse_check_tol: float = 2e-8,
        reverse_check_norm: NormFunction = maximum_norm,
        projection_solver: ProjectionSolver = solve_projection_onto_manifold_newton,
        projection_solver_kwargs: dict[str, Any] | None = None,
    ) -> None:
        """
        Args:
            system: Hamiltonian system to integrate the constrained dynamics of.
            step_size: Integrator time step. If set to :code:`None` it is assumed that a
                step size adapter will be used to set the step size before calling the
                :py:meth:`step` method.
            n_inner_step: Positive integer specifying number of 'inner' constrained
                :code:`system.h2_flow` steps to take within each overall step. As the
                derivative :code:`system.dh1_dpos` is not evaluated during the
                :code:`system.h2_flow` steps, if this derivative is relatively expensive
                to compute compared to evaluating :code:`system.h2_flow` then compared
                to using :code:`n_inner_step = 1` (the default) for a given
                :code:`step_size` it can be more computationally efficient to use
                :code:`n_inner_step > 1` in combination within a larger
                :code:`step_size`, thus reducing the number of :code:`system.dh1_dpos`
                evaluations to simulate forward a given time while still controlling the
                effective time step used for the constrained :code:`system.h2_flow`
                steps which involve solving a non-linear system of equations to retract
                the position component of the updated state back on to the manifold,
                with the iterative solver typically diverging if the
                time step used is too large.
            reverse_check_tol: Tolerance for check of reversibility of implicit
                sub-steps which involve iterative solving of a non-linear system of
                equations. The step is assumed to be reversible if sequentially applying
                the forward and adjoint updates to a state returns to a state with a
                position component within a distance (defined by the
                `reverse_check_norm` argument) of :code:`reverse_check_tol` of the
                original state position component. If this condition is not met a
                :py:exc:`mici.errors.NonReversibleStepError` exception is raised.
            reverse_check_norm: Norm function accepting a single one-dimensional array
                input and returning a non-negative floating point value defining the
                distance to use in the reversibility check. Defaults to
                :py:func:`mici.solvers.maximum_norm`.
            projection_solver: Function which given two states :code:`state` and
                :code:`state_prev`, floating point time step :code:`time_step` and a
                Hamiltonian system object :code:`system` solves the non-linear system of
                equations in :code:`λ`.

                .. code::

                    system.constr(
                        state.pos
                        + dh2_flow_pos_dmom @ system.jacob_constr(state_prev).T @ λ
                    ) == 0

                where :code:`dh2_flow_pos_dmom = system.dh2_flow_dmom(time_step)[0]` is
                the derivative of the action of the (linear) :code:`system.h2_flow` map
                on the state momentum component with respect to the position component.
                This is used to project the state position component back on to the
                manifold after an unconstrained :code:`system.h2_flow` update. If the
                solver fails to convege a :py:exc:`mici.errors.ConvergenceError`
                exception is raised.
            projection_solver_kwargs: Dictionary of any keyword arguments to
                :code:`projection_solver`.
        """
        super().__init__(system, step_size)
        self.n_inner_step = n_inner_step
        self.reverse_check_tol = reverse_check_tol
        self.reverse_check_norm = reverse_check_norm
        self.projection_solver = projection_solver
        if projection_solver_kwargs is None:
            projection_solver_kwargs = {}
        self.projection_solver_kwargs = projection_solver_kwargs

    def _h2_flow_retraction_onto_manifold(
        self,
        state: ChainState,
        state_prev: ChainState,
        time_step: float,
    ) -> None:
        self.system.h2_flow(state, time_step)
        self.projection_solver(
            state,
            state_prev,
            time_step,
            self.system,
            **self.projection_solver_kwargs,
        )

    def _project_onto_cotangent_space(self, state: ChainState) -> None:
        state.mom = self.system.project_onto_cotangent_space(state.mom, state)

    def _step_a(self, state: ChainState, time_step: float) -> None:
        self.system.h1_flow(state, time_step)
        self._project_onto_cotangent_space(state)

    def _step_b(self, state: ChainState, time_step: float) -> None:
        time_step_inner = time_step / self.n_inner_step
        for i in range(self.n_inner_step):
            state_prev = state.copy()
            self._h2_flow_retraction_onto_manifold(state, state_prev, time_step_inner)
            if i == self.n_inner_step - 1:
                # If at last inner step pre-evaluate dh1_dpos before projecting
                # state on to cotangent space, with computed value being
                # cached. During projection the constraint Jacobian at new
                # position will be calculated however if we are going to make a
                # h1_flow step immediately after we will evaluate dh1_dpos
                # which may involve evaluating the gradient of the log
                # determinant of the Gram matrix, during which we will evaluate
                # the constraint Jacobian in the forward pass anyway.
                # Pre-evaluating here therefore saves one extra Jacobian
                # evaluation when the target density includes a Gram matrix log
                # determinant term (and will not add any cost if this is not
                # the case as dh1_dpos will still be cached and reused).
                self.system.dh1_dpos(state)
            self._project_onto_cotangent_space(state)
            state_back = state.copy()
            self._h2_flow_retraction_onto_manifold(state_back, state, -time_step_inner)
            rev_diff = self.reverse_check_norm(state_back.pos - state_prev.pos)
            if rev_diff > self.reverse_check_tol:
                msg = (
                    f"Non-reversible step. Distance between initial and "
                    f"forward-backward integrated positions = {rev_diff:.1e}."
                )
                raise NonReversibleStepError(msg)

    def _step(self, state: ChainState, time_step: float) -> None:
        self._step_a(state, 0.5 * time_step)
        self._step_b(state, time_step)
        self._step_a(state, 0.5 * time_step)



"""Classes for staging sampling of Markov chains."""

from __future__ import annotations

import abc
from typing import TYPE_CHECKING, NamedTuple

if TYPE_CHECKING:
    from collections.abc import Iterable

    from mici.adapters import Adapter
    from mici.types import TraceFunction


class ChainStage(NamedTuple):
    """Parameters of chain stage.

    Parameters:
        n_iter: Number of iterations in chain stage.
        adapters: Dictionary of adapters to apply to each transition in chain stage.
        trace_funcs: Functions defining chain variables to trace during chain stage.
        record_stats: Whether to record statistics during chain stage.
    """

    n_iter: int
    adapters: dict[str, Iterable[Adapter]]
    trace_funcs: tuple[TraceFunction] | None
    record_stats: bool


class Stager(abc.ABC):
    """Abstract chain iteration stager."""

    @abc.abstractmethod
    def stages(
        self,
        n_warm_up_iter: int,
        n_main_iter: int,
        adapters: dict[str, Iterable[Adapter]],
        trace_funcs: Iterable[TraceFunction],
        *,
        trace_warm_up: bool = False,
    ) -> dict[str, ChainStage]:
        """Create dictionary specifying labels and parameters of chain sampling stages.

        Constructs an ordered dictionary with entries corresponding to the sequence of
        sampling stages when running chains with one or more initial adaptation stages.
        The keys of each entry are a string label for the sampling stage and the value a
        :py:class:`ChainStage` named tuple specifying the parameters of the stage.

        Args:
            n_warm_up_iter: Number of adaptive warm up iterations per chain. Depending
                on the adapters specified by the `adapters` argument the warm up
                iterations may be split between one or more adaptive stages.
            n_main_iter: Number of iterations (samples to draw) per chain during main
                (non-adaptive) sampling stage.
            trace_funcs: Iterable of functions which compute the variables to be
                recorded at each chain iteration, with each trace function being passed
                the current state and returning a dictionary of scalar or array values
                corresponding to the variable(s) to be stored. The keys in the returned
                dictionaries are used to index the trace arrays in the returned traces
                dictionary. If a key appears in multiple dictionaries only the the value
                corresponding to the last trace function to return that key will be
                stored. By default chains are only traced in the iterations of the final
                non-adaptive stage - this behaviour can be altered using the
                `trace_warm_up` argument.
            adapters: Dictionary of iterables of :py:class:`mici.adapters.Adapter`
                instances keyed by strings corresponding to the key of the transition in
                the sampler
                :py:attr:`mici.samplers.MarkovChainMonteCarloMethod.transitions`
                dictionary to apply the adapters to, to use to adaptatively set
                parameters of the transitions during the adaptive stages of the chains.
                Note that the adapter updates are applied in the order the adapters
                appear in the iterables and so if multiple adapters change the same
                parameter(s) the order will matter.
            trace_warm_up: Whether to record chain traces and statistics during warm-up
                stage iterations (:code:`True`) or only record traces and statistics in
                the iterations of the final non-adaptive stage (:code:`False`, the
                default).

        Returns:
            Dictionary specifying chain stages, keyed by labels for stages.
        """


class WarmUpStager(Stager):
    """Chain iteration stager with a single adaptive warm up stage.

    Sampling is split in to two stages:

      1. An adaptive warm up stage will all adapters active.
      2. A main sampling stage with no adapters active.

    Only in the main sampling stage are traces of the chain state recorded by storing
    the outputs of functions of the sampled chain state after each iteration.
    """

    def stages(
        self,
        n_warm_up_iter: int,
        n_main_iter: int,
        adapters: dict[str, Iterable[Adapter]],
        trace_funcs: Iterable[TraceFunction],
        *,
        trace_warm_up: bool = False,
    ) -> dict[str, ChainStage]:
        sampling_stages = {}
        trace_funcs = tuple(trace_funcs) if trace_funcs is not None else trace_funcs
        # adaptive warm up stage
        if n_warm_up_iter > 0:
            warm_up_trace_funcs = trace_funcs if trace_warm_up else None
            sampling_stages["Adaptive warm up"] = ChainStage(
                n_iter=n_warm_up_iter,
                adapters=adapters,
                trace_funcs=warm_up_trace_funcs,
                record_stats=trace_warm_up,
            )
        # main non-adaptive stage
        if n_main_iter > 0:
            sampling_stages["Main non-adaptive"] = ChainStage(
                n_iter=n_main_iter,
                adapters=None,
                trace_funcs=trace_funcs,
                record_stats=True,
            )
        return sampling_stages


class WindowedWarmUpStager(Stager):
    """Chain iteration stager with a hierarchy of adaptive warm up stages.

    Following the approach of [Stan](https://mc-stan.org) the adaptive stages are split
    in to two types - 'fast' adaptation stages which adjust only transition parameters
    which can be adapted quickly using local information and 'slow' adaptation stages
    which *addtionally* adjust transition parameters which require more global
    information. The adapters to be used in both the fast and slow adaptation stages
    will be referred to as the *fast adapters* and the adapters to use in only the slow
    adaptation stages the *slow adapters*. Each adapter self identifies if it is a fast
    adapter by whether the :py:attr:`mici.adapters.Adapter.is_fast` attribute is set to
    :code:`True`.

    The adaptive warm up iterations are split into three stages:

      1. An initial fast adaptive stage with only fast adapters active.
      2. A slow adaptive stage with both slow and fast adapters active.
      3. A final adaptive stage with only fast adapters active.

    The slow sampling stage (2) is further split in to a sequence of growing, memoryless
    windows with the adapter stages reset at the beginning of each window, and the
    number of iterations in each window increasing (by default doubling). The split of
    the iterations in each of these stages can be controlled using the keyword arguments
    `n_init_fast_stage_iter`, `n_init_slow_window_iter`, `n_final_fast_stage_iter` and
    `slow_window_multiplier` (see descriptions below).

    After the initial adaptive warm up stages a subsequent main sampling stage with no
    further adaptation is performed. Only in this main sampling stage are traces of the
    chain state recorded by storing the outputs of functions of the sampled chain state
    after each iteration.
    """

    def __init__(
        self,
        n_init_slow_window_iter: int = 25,
        n_init_fast_stage_iter: int = 75,
        n_final_fast_stage_iter: int = 50,
        slow_window_multiplier: float = 2.0,
    ) -> None:
        """
        Args:
            n_init_slow_window_iter: Number of iterations in the initial (smallest)
                window in the slow adaptation stage. Defaults to 25. If the sum of
                `n_init_slow_window_iter`, `n_init_fast_stage_iter` and
                `n_final_fast_stage_iter` is more than `n_warm_up_iter` then
                `n_init_slow_window_iter` is set to approximately 75% of
                `n_warm_up_iter` (with a single window being used in the slow adaptation
                stage in this case).
            n_init_fast_stage_iter: Number of iterations in the initial fast adaptation
                stage. Defaults to 75. If the sum of `n_init_slow_window_iter`,
                n_init_fast_stage_iter` and `n_final_fast_stage_iter` is more than
                `n_warm_up_iter` then `n_init_fast_stage_iter` is set to approximately
                15% of `n_warm_up_iter`.
            n_final_fast_stage_iter: Number of iterations in the final fast adaptation
                stage. Defaults to 50. If the sum of `n_init_slow_window_iter`,
                `n_init_fast_stage_iter` and `n_final_fast_stage_iter` is more than
                `n_warm_up_iter` then `n_init_fast_stage_iter` is set to approximately
                10% of `n_warm_up_iter`.
            slow_window_multiplier: Multiplier by which to increase the number of
                iterations of each subsequent slow adaptation window by. Defaults to 2
                such that each window doubles in size.
        """
        self.n_init_slow_window_iter = n_init_slow_window_iter
        self.n_init_fast_stage_iter = n_init_fast_stage_iter
        self.n_final_fast_stage_iter = n_final_fast_stage_iter
        self.slow_window_multiplier = slow_window_multiplier

    def stages(
        self,
        n_warm_up_iter: int,
        n_main_iter: int,
        adapters: dict[str, Iterable[Adapter]],
        trace_funcs: Iterable[TraceFunction],
        *,
        trace_warm_up: bool = False,
    ) -> dict[str, ChainStage]:
        trace_funcs = tuple(trace_funcs) if trace_funcs is not None else trace_funcs
        fast_adapters = {
            trans_key: [adapter for adapter in adapter_list if adapter.is_fast]
            for trans_key, adapter_list in adapters.items()
        }
        if (
            self.n_init_fast_stage_iter
            + self.n_init_slow_window_iter
            + self.n_final_fast_stage_iter
        ) > n_warm_up_iter:
            n_init_fast_stage_iter = int(0.15 * n_warm_up_iter)
            n_final_fast_stage_iter = int(0.1 * n_warm_up_iter)
            n_init_slow_window_iter = (
                n_warm_up_iter - n_init_fast_stage_iter - n_final_fast_stage_iter
            )
        else:
            n_init_slow_window_iter = self.n_init_slow_window_iter
            n_init_fast_stage_iter = self.n_init_fast_stage_iter
            n_final_fast_stage_iter = self.n_final_fast_stage_iter
        sampling_stages = {}
        # adaptive warm-up stages
        if n_warm_up_iter > 0:
            warm_up_trace_funcs = trace_funcs if trace_warm_up else None
            record_stats = trace_warm_up
            # initial fast adaptation stage
            sampling_stages["Initial fast adaptive"] = ChainStage(
                n_iter=n_init_fast_stage_iter,
                adapters=fast_adapters,
                trace_funcs=warm_up_trace_funcs,
                record_stats=record_stats,
            )
            # growing size slow adaptation windows
            n_window_iter = n_init_slow_window_iter
            slow_windows = []
            counter = 0
            n_slow_stage_iter = (
                n_warm_up_iter - n_init_fast_stage_iter - n_final_fast_stage_iter
            )
            while counter < n_slow_stage_iter:
                # check if iteration counter at end of next loop iteration will be
                # greater than total number of warm up iterations and if so set number
                # of iterations in current window to be equal to all remaining warm up
                # iterations
                counter_next = counter + int(
                    (1 + self.slow_window_multiplier) * n_window_iter,
                )
                if counter_next > n_slow_stage_iter:
                    n_window_iter = n_slow_stage_iter - counter
                slow_windows.append(n_window_iter)
                counter += n_window_iter
                n_window_iter = int(self.slow_window_multiplier * n_window_iter)
            for i, n_iter in enumerate(slow_windows):
                sampling_stages[f"Slow adaptive ({i + 1}/{len(slow_windows)})"] = (
                    ChainStage(
                        n_iter=n_iter,
                        adapters=adapters,
                        trace_funcs=warm_up_trace_funcs,
                        record_stats=record_stats,
                    )
                )
            # final fast adaptation stage
            sampling_stages["Final fast adaptive"] = ChainStage(
                n_iter=n_final_fast_stage_iter,
                adapters=fast_adapters,
                trace_funcs=warm_up_trace_funcs,
                record_stats=record_stats,
            )
        # main non-adaptive stage
        if n_main_iter > 0:
            sampling_stages["Main non-adaptive"] = ChainStage(
                n_iter=n_main_iter,
                adapters=None,
                trace_funcs=trace_funcs,
                record_stats=True,
            )
        return sampling_stages



"""Utilities for interfacing with external probabilistic programming libraries."""

from __future__ import annotations

import importlib
import os
from typing import TYPE_CHECKING

import numpy as np

import mici

if TYPE_CHECKING:
    from typing import Literal

    import arviz
    import pymc
    import stan

    from mici.states import ChainState
    from mici.types import (
        ArrayLike,
        GradientFunction,
        ScalarFunction,
        ScalarLike,
        TraceFunction,
    )


def convert_to_inference_data(
    traces: dict[str, list[ArrayLike]],
    stats: dict[str, list[ArrayLike]],
    energy_key: str | None = "energy",
    lp_key: str | None = "lp",
) -> arviz.InferenceData:
    """Convert Mici :code:`sample_chains` output to :py:class:`arviz.InferenceData`.

    Args:
        traces: Traces output from Mici
            :py:meth:`mici.samplers.MarkovChainMonteCarloMethod.sample_chains` call. A
            dictionary of variables traced over sampled chains with the dictionary keys
            the variable names and the values a list of arrays, one array per sampled
            chain, with the first array dimension corresponding to the draw index and
            any remaining dimensions, the variable dimensions.
        stats: Statistics output from Mici `sample_chains` call. A dictionary of chain
            statistics traced over sampled chains with the dictionary keys the
            statistics names and the values a list of arrays, one array per sampled
            chain, with the array dimension corresponding to the draw index.
        energy_key: The key of an entry in the `traces` dictionary corresponding the
            value of the Hamiltonian energy for the accepted proposal (up to an additive
            constant). If present the corresponding values will be added to the
            `sample_stats` group of the returned `InferenceData` object.
        lp_key: The key of an entry in the `traces` dictionary corresponding the value
            of the joint log posterior density for the model (up to an additive
            constant). If present the corresponding values will be added to the
            `sample_stats` group of the returned `InferenceData` object.

    Returns:
        ArviZ inference data object with traced chain data stored in the `posterior`
        group and additional chain statistics in the `sample_stats` group.
    """
    import arviz

    stats = stats.copy()
    stats["n_steps"] = stats.pop("n_step")
    stats["acceptance_rate"] = stats.pop("accept_stat")
    if energy_key is not None and energy_key in traces:
        stats["energy"] = traces[energy_key]
    if lp_key is not None and lp_key in traces:
        stats["lp"] = traces[lp_key]
    return arviz.InferenceData(
        posterior=arviz.dict_to_dataset(traces, library=mici),
        sample_stats=arviz.dict_to_dataset(stats, library=mici),
    )


def construct_pymc_model_functions(
    model: pymc.Model,
) -> tuple[ScalarFunction, GradientFunction, TraceFunction]:
    """Construct functions for sampling from PyMC model using Mici.

    Args:
        model: PyMC model to construct functions for.

    Returns:
        Tuple :code:`(neg_log_dens, grad_neg_log_dens, trace_func)` with
        :code:`neg_log_dens` a function for evaluating negative logarithm of
        unnormalized posterior density associated with model, :code:`grad_neg_log_dens`
        a function for evaluating gradient of :code:`neg_log_dens` with respect to
        position array argument and :code:`trace_func` a function which extract model
        parameter values from chain state for tracing during sampling.
    """
    import pymc

    initial_point = model.initial_point()
    raveled_initial_point = pymc.blocking.DictToArrayBijection.map(initial_point)

    val_and_grad_log_dens = model.logp_dlogp_function()
    val_and_grad_log_dens.set_extra_values({})

    def grad_neg_log_dens(pos: ArrayLike) -> tuple[ArrayLike, ScalarLike]:
        val, grad = val_and_grad_log_dens(pos)
        return -grad, -val

    def neg_log_dens(pos: ArrayLike) -> ScalarLike:
        val, _ = val_and_grad_log_dens(pos)
        return -val

    def trace_func(state: ChainState) -> dict[str, ScalarLike]:
        raveled_vars = pymc.blocking.RaveledVars(
            state.pos,
            raveled_initial_point.point_map_info,
        )
        var_dict = pymc.blocking.DictToArrayBijection.rmap(raveled_vars)
        trace_dict = {}
        for rv in model.unobserved_RVs:
            if rv.name in var_dict:
                trace_dict[rv.name] = var_dict[rv.name]
            else:
                transform = model.rvs_to_transforms[rv]
                trace_dict[rv.name] = transform.backward(
                    var_dict[f"{rv.name}_{transform.name}__"],
                    *rv.owner.inputs,
                ).eval()
        trace_dict["lp"] = -neg_log_dens(state.pos)
        return trace_dict

    return neg_log_dens, grad_neg_log_dens, trace_func


def sample_pymc_model(
    draws: int = 1000,
    *,
    tune: int = 1000,
    chains: int | None = None,
    cores: int | None = None,
    random_seed: int | None = None,
    progressbar: bool = True,
    init: Literal["auto", "adapt_diag", "jitter+adapt_diag", "adapt_full"] = "auto",
    jitter_max_retries: int = 10,
    return_inferencedata: bool = False,
    model: pymc.Model | None = None,
    target_accept: float = 0.8,
    max_treedepth: int = 10,
) -> arviz.InferenceData | dict[str, ArrayLike]:
    """Generate approximate samples from posterior defined by a PyMC model.

    Uses dynamic multinomial HMC algorithm in Mici with adaptive warm-up phase.

    This function replicates the interface of the :py:func:`pymc.sample` function to
    allow using as a (partial) drop-in replacement.

    Args:
        draws: The number of samples to draw.
        tune: Number of iterations to tune, defaults to 1000. Samplers adjust the step
            sizes, scalings or similar during tuning. Tuning samples will be drawn in
            addition to the number specified in the `draws` argument, and will be
            discarded.
        chains: The number of chains to sample. Running independent chains is important
            for some convergence statistics and can also reveal multiple modes in the
            posterior. If :code::`None`, then set to either :code:`cores` or 2,
            whichever is larger.
        cores: The number of chains to run in parallel. If :code:`None`, set to the
            number of CPU cores in the system, but at most 4.
        random_seed: Seed for NumPy random number generator used for generating random
            variables while sampling chains. If :code:`None` then generator will be
            seeded with entropy from operating system.
        progressbar: Whether or not to display a progress bar.
        init: Initialization method to use. One of:

            * :code:`"adapt_diag"`: Start with a identity mass matrix and then adapt a
              diagonal based on the variance of the tuning samples. All chains use the
              test value (usually the prior mean) as starting point.
            * :code:`jitter+adapt_diag`: Same as :code:`"adapt_diag"`, but add uniform
              jitter in [-1, 1] to the starting point in each chain. Also chosen if
              :code:`init="auto"`.
            * :code:`"adapt_full"`: Adapt a dense mass matrix using the sample
              covariances.
            * :code:`jitter+adapt_full`: Same as :code:`"adapt_full"`, but add uniform
              jitter in [-1, 1] to the starting point in each chain.d

        jitter_max_retries: Maximum number of repeated attempts (per chain) at creating
            an initial matrix with uniform jitter that yields a finite probability. This
            applies to `"jitter+adapt_diag"` and :code:`"jitter+adapt_full"`
            :py:obj:`init` methods.
        return_inferencedata: Whether to return the traces as an
            :py:class:`arviz.InferenceData` (:code:`True`) object or a
            :py:class:`dict` (:code:`False`).
        model: PyMC model defining posterior distribution to sample from. May be
            :code:`None` if function is called from within model context manager.
        target_accept: Target value for the acceptance statistic being controlled during
            adaptive warm-up.
        max_treedepth: Maximum depth to expand trajectory binary tree to in integrator
            transition. The maximum number of integrator steps corresponds to
            :code:`2**max_treedepth`.

    Returns:
        A dictionary or :py:class:`arviz.InferenceData` object containing the sampled
        chain output. Dictionary output (when :code:`return_inferencedata=False`) has
        string keys corresponding to the name of each traced variable in the model, with
        the values being the corresponding values of the variables traced across the
        chains as NumPy arrays, with the first dimension the chain index  (of size equal
        to :code:`chains`), the second dimension the draw index (of size equal to
        :code:`draws`) and any remaining dimensions corresponding to the dimensions of
        the traced variable. If :code:`return_inferencedata=True` an
        :py:class:`arviz.InferenceData` object is instead returned with traced chain
        data stored in the :code:`posterior` group and additional chain statistics in
        the :code:`sample_stats` group.
    """
    import pymc

    if return_inferencedata and importlib.util.find_spec("arviz") is None:
        msg = "Cannot return InferenceData as ArviZ is not installed"
        raise ValueError(msg)

    model = pymc.modelcontext(model)

    # assume 2 threads per CPU core
    cores = min(4, os.cpu_count() // 2) if cores is None else cores
    chains = max(2, cores) if chains is None else chains

    init = "jitter+adapt_diag" if init == "auto" else init
    if init in ("jitter+adapt_diag", "jitter+adapt_full", "adapt_diag", "adapt_full"):
        use_dense_metric = "adapt_full" in init
        jitter_init = "jitter" in init
    else:
        msg = 'init must be "auto", "jitter+adapt_diag", "adapt_diag" or "adapt_full"'
        raise ValueError(msg)

    neg_log_dens, grad_neg_log_dens, trace_func = construct_pymc_model_functions(model)

    system = mici.systems.EuclideanMetricSystem(
        neg_log_dens=neg_log_dens,
        grad_neg_log_dens=grad_neg_log_dens,
        backend=None,
    )

    integrator = mici.integrators.LeapfrogIntegrator(system)
    rng = np.random.default_rng(random_seed)
    sampler = mici.samplers.DynamicMultinomialHMC(
        system,
        integrator,
        rng,
        max_tree_depth=max_treedepth,
    )

    step_size_adapter = mici.adapters.DualAveragingStepSizeAdapter(target_accept)
    metric_adapter = (
        mici.adapters.OnlineCovarianceMetricAdapter()
        if use_dense_metric
        else mici.adapters.OnlineVarianceMetricAdapter()
    )

    initial_point = model.initial_point()
    raveled_initial_point = pymc.blocking.DictToArrayBijection.map(initial_point)

    if jitter_init:
        mean = raveled_initial_point.data.copy()
        init_states = []
        for _c in range(chains):
            for _t in range(jitter_max_retries):
                pos = mean + rng.uniform(-1, 1, mean.shape)
                if np.isfinite(neg_log_dens(pos)):
                    break
            init_states.append(pos)
    else:
        init_states = [raveled_initial_point.data.copy() for c in range(chains)]

    _, traces, stats = sampler.sample_chains(
        n_warm_up_iter=tune,
        n_main_iter=draws,
        init_states=init_states,
        adapters=[step_size_adapter, metric_adapter],
        trace_funcs=[trace_func],
        n_process=cores,
        display_progress=progressbar,
        monitor_stats=["accept_stat", "n_step", "diverging"],
    )

    if return_inferencedata:
        return convert_to_inference_data(traces, stats)
    return {k: np.stack(v) for k, v in traces.items()}


def get_stan_model_unconstrained_param_dim(model: stan.Model) -> int:
    """Get total dimension of unconstrained parameters in Stan model.

    Args:
        model: Stan model to get dimension for.

    Returns:
        Non-negative integer specifying unconstrained parameter dimension.
    """
    param_size_list = [np.prod(dim, dtype=np.int64) for dim in model.dims]
    n_dim = sum(param_size_list)
    while True:
        try:
            model.log_prob([0] * n_dim)
        except RuntimeError:  # noqa: PERF203
            param_size_list.pop()
            n_dim = sum(param_size_list)
        else:
            return n_dim


def construct_stan_model_functions(
    model: stan.Model,
) -> tuple[ScalarFunction, GradientFunction, TraceFunction]:
    """Construct functions for sampling from Stan model using Mici.

    Args:
        model: Stan model to construct functions for.

    Returns:
        Tuple :code:`(neg_log_dens, grad_neg_log_dens, trace_func)` with
        :code:`neg_log_dens` a function for evaluating negative logarithm of
        unnormalized posterior density associated with model, :code:`grad_neg_log_dens`
        a function for evaluating gradient of :code:`neg_log_dens` with respect to
        position array argument and :code:`trace_func` a function which extract model
        parameter values from chain state for tracing during sampling.
    """

    def neg_log_dens(u: ArrayLike) -> ScalarLike:
        return -model.log_prob(list(u))

    def grad_neg_log_dens(u: ArrayLike) -> ArrayLike:
        return -np.array(model.grad_log_prob(list(u)))

    param_size_list = [np.prod(dim, dtype=np.int64) for dim in model.dims]

    def trace_func(state: ChainState) -> dict[str, ScalarLike]:
        param_array = np.array(model.constrain_pars(list(state.pos)))
        trace_dict = {
            name: val.reshape(shape)
            for name, val, shape in zip(
                model.param_names,
                np.split(param_array, np.cumsum(param_size_list)[:-1]),
                model.dims,
                strict=True,
            )
        }
        trace_dict["lp"] = -neg_log_dens(state.pos)
        return trace_dict

    return neg_log_dens, grad_neg_log_dens, trace_func


def sample_stan_model(
    model_code: str,
    data: dict,
    *,
    num_samples: int = 1000,
    num_warmup: int = 1000,
    num_chains: int = 4,
    save_warmup: bool = False,
    metric: Literal["unit_e", "diag_e", "dense_e"] = "diag_e",
    stepsize: float = 1.0,
    adapt_engaged: bool = True,
    delta: float = 0.8,
    gamma: float = 0.05,
    kappa: float = 0.75,
    t0: int = 10,
    init_buffer: int = 75,
    term_buffer: int = 50,
    window: int = 25,
    max_depth: int = 10,
    seed: int | None = None,
    return_inferencedata: bool = False,
) -> arviz.InferenceData | dict[str, ArrayLike]:
    """Generate approximate samples from posterior defined by a Stan model.

    Uses dynamic multinomial HMC algorithm in Mici with adaptive warm-up phase.

    This function follows a similar argument naming scheme to the PyStan
    :py:meth:`stan.model.Model.sample` method (which itself follows CmdStan) to allow
    using as a (partial) drop-in replacement.

    Args:
        model_code: Stan program code describing a Stan model.
        data: A Python dictionary or mapping providing the data for the model. Variable
            names are the keys and the values are their associated values.
        num_samples: A non-negative integer specifying the number of non-warm-up
            iterations per chain.
        num_warmup: A non-negative integer specifying the number of warm-up iterations
            per chain.
        num_chains: A positive integer specifying the number of Markov chains.
        save_warmup: Whether to save warm-up chain data (`True`) or not (`False`).
        metric: String specifying metric type. One of "unit_e", "diag_e" or "dense_e",
            indicating respectively to used a fixed identity matrix metric
            representation, to use a diagonal metric matrix representation adapted based
            on estimates of the marginal posterior variances, to use a dense metric
            matrix representation based on estimates of the posterior covariance matrix.
        stepsize: Initial integrator step size.
        adapt_engaged: Whether adaptation is engaged (`True`) or not (`False`).
        delta: Adaptation target acceptance statistic.
        gamma: Adaptation regularization scale.
        kappa: Adaptation relaxation exponent.
        t0: Adaptation iteration offset.
        init_buffer: Width of initial fast adaptation interval.
        term_buffer: Width of final fast adaptation interval.
        window: Initial width of slow adaptation interval.
        max_depth: Maximum depth of binary trajectory tree.
        seed: Seed for Numpy random number generator used for generating random
            variables while sampling chains. If `None` then generator will be seeded
            with entropy from operating system.
        return_inferencedata: Whether to return the traces as an `arviz.InferenceData`
            (`True`) object or a dict (`False`).

    Returns:
        A dictionary or ArviZ `InferenceData` object containing the sampled chain
        output. Dictionary output (when `return_inferencedata=False`) has string keys
        corresponding to the name of each traced variable in the model, with the values
        being the corresponding values of the variables traced across the chains as
        NumPy arrays, with the first dimension the flattened draw index across all
        chains (of size equal to `num_chains * num_samples`) and any remaining
        dimensions corresponding to the dimensions of the traced variable. If
        `return_inferencedata=True` an ArviZ `InferenceData` object is instead returned
        with traced chain data stored in the `posterior` group and additional chain
        statistics in the `sample_stats` group.
    """
    import stan

    if return_inferencedata and importlib.util.find_spec("arviz") is None:
        msg = "Cannot return InferenceData as ArviZ is not installed"
        raise ValueError(msg)

    model = stan.build(model_code, data=data)

    neg_log_dens, grad_neg_log_dens, trace_func = construct_stan_model_functions(model)

    system = mici.systems.EuclideanMetricSystem(
        neg_log_dens=neg_log_dens,
        grad_neg_log_dens=grad_neg_log_dens,
        backend=None,
    )

    integrator = mici.integrators.LeapfrogIntegrator(system, step_size=stepsize)
    rng = np.random.default_rng(seed)
    sampler = mici.samplers.DynamicMultinomialHMC(
        system,
        integrator,
        rng,
        max_tree_depth=max_depth,
    )

    if adapt_engaged:
        step_size_adapter = mici.adapters.DualAveragingStepSizeAdapter(
            adapt_stat_target=delta,
            iter_offset=t0,
            iter_decay_coeff=kappa,
            log_step_size_reg_coefficient=gamma,
        )
        adapters = [step_size_adapter]

        if metric == "diag_e":
            adapters.append(mici.adapters.OnlineVarianceMetricAdapter())
        elif metric == "dense_e":
            adapters.append(mici.adapters.OnlineCovarianceMetricAdapter())

        if len(adapters) > 1:
            stager = mici.stagers.WindowedWarmUpStager(
                n_init_fast_stage_iter=init_buffer,
                n_final_fast_stage_iter=term_buffer,
                n_init_slow_window_iter=window,
            )
        else:
            stager = mici.stagers.WarmUpStager()
    else:
        adapters = None
        stager = None

    dim_u = get_stan_model_unconstrained_param_dim(model)
    init_states = rng.uniform(-2, 2, size=(num_chains, dim_u))

    _, traces, stats = sampler.sample_chains(
        n_warm_up_iter=num_warmup,
        n_main_iter=num_samples,
        init_states=init_states,
        adapters=adapters,
        stager=stager,
        trace_funcs=[trace_func],
        monitor_stats=["accept_stat", "n_step", "diverging"],
        trace_warm_up=save_warmup,
    )

    if return_inferencedata:
        return convert_to_inference_data(traces, stats)
    return {k: np.concatenate(v).swapaxes(0, -1) for k, v in traces.items()}



import numpy as np
import pytest

import mici

SEED = 3046987125


@pytest.fixture
def rng():
    return np.random.default_rng(SEED)


@pytest.fixture(params=(1, 3, (3, 2), (5, 5, 1), (10, 10, 10)))
def array_equal_pair(rng, request):
    array = rng.standard_normal(request.param)
    return array, array.copy()


def test_hash_array_equal(array_equal_pair):
    array_1, array_2 = array_equal_pair
    assert mici.utils.hash_array(array_1) == mici.utils.hash_array(array_2)


@pytest.fixture(params=((1, 1), (1, 3), ((5, 5), (5, 5)), ((5, 5, 1), (3, 2))))
def array_not_equal_pair(rng, request):
    shape_1, shape_2 = request.param
    return rng.standard_normal(shape_1), rng.standard_normal(shape_2)


def test_hash_array_not_equal(array_not_equal_pair):
    array_1, array_2 = array_not_equal_pair
    assert mici.utils.hash_array(array_1) != mici.utils.hash_array(array_2)


def get_val(obj):
    if isinstance(obj, mici.utils.LogRepFloat):
        return obj.val
    return obj


class TestLogRepFloat:
    VALS = sorted((0.0, 0.9, 1, 1.1, 2.1, 120.0))
    FIXED = VALS[1]

    @pytest.fixture(params=VALS)
    def val_pair(self, request):
        val = request.param
        return mici.utils.LogRepFloat(val), val

    @pytest.fixture(params=VALS)
    def other_pair(self, request):
        val = request.param
        return mici.utils.LogRepFloat(val), val

    def test_greater_than(self, val_pair, other_pair):
        if val_pair[1] != other_pair[1]:
            assert (val_pair[0] > other_pair[0]) == (val_pair[1] > other_pair[1])
            assert (val_pair[0] > other_pair[1]) == (val_pair[1] > other_pair[0])

    def test_less_than(self, val_pair, other_pair):
        if val_pair[1] != other_pair[1]:
            assert (val_pair[0] < other_pair[0]) == (val_pair[1] < other_pair[1])
            assert (val_pair[0] < other_pair[1]) == (val_pair[1] < other_pair[0])

    def test_greater_than_or_equal(self, val_pair, other_pair):
        assert val_pair[0] >= val_pair[0]
        if val_pair[1] != other_pair[1]:
            assert (val_pair[0] >= other_pair[0]) == (val_pair[1] >= other_pair[1])
            assert (val_pair[0] >= other_pair[1]) == (val_pair[1] >= other_pair[0])

    def test_less_than_or_equal(self, val_pair, other_pair):
        assert val_pair[0] <= val_pair[0]
        if val_pair[1] != other_pair[1]:
            assert (val_pair[0] <= other_pair[0]) == (val_pair[1] <= other_pair[1])
            assert (val_pair[0] <= other_pair[1]) == (val_pair[1] <= other_pair[0])

    def test_equal_to(self, val_pair, other_pair):
        assert val_pair[0] == val_pair[0]
        if val_pair[1] != other_pair[1]:
            assert (val_pair[0] == other_pair[0]) == (val_pair[1] == other_pair[1])
            assert (val_pair[0] == other_pair[1]) == (val_pair[1] == other_pair[0])

    def test_not_equal_to(self, val_pair, other_pair):
        assert val_pair[0] == val_pair[0]
        if val_pair[1] != other_pair[1]:
            assert (val_pair[0] != other_pair[0]) == (val_pair[1] != other_pair[1])
            assert (val_pair[0] != other_pair[1]) == (val_pair[1] != other_pair[0])

    def test_mult(self, val_pair, other_pair):
        assert np.isclose(
            get_val(val_pair[0] * other_pair[0]),
            get_val(val_pair[1] * other_pair[1]),
        )
        assert np.isclose(
            get_val(val_pair[0] * other_pair[1]),
            get_val(val_pair[1] * other_pair[0]),
        )

    def test_div(self, val_pair, other_pair):
        if other_pair[1] != 0:
            assert np.isclose(
                get_val(val_pair[0] / other_pair[0]),
                get_val(val_pair[1] / other_pair[1]),
            )
            assert np.isclose(
                get_val(val_pair[0] / other_pair[1]),
                get_val(val_pair[1] / other_pair[0]),
            )

    def test_add(self, val_pair, other_pair):
        assert np.isclose(
            get_val(val_pair[0] + other_pair[0]),
            get_val(val_pair[1] + other_pair[1]),
        )
        assert np.isclose(
            get_val(val_pair[0] + other_pair[1]),
            get_val(val_pair[1] + other_pair[0]),
        )

    def test_sub(self, val_pair, other_pair):
        assert np.isclose(
            get_val(val_pair[0] - other_pair[0]),
            get_val(val_pair[1] - other_pair[1]),
        )
        assert np.isclose(
            get_val(val_pair[0] - other_pair[1]),
            get_val(val_pair[1] - other_pair[0]),
        )

    def test_neg(self, val_pair):
        assert np.isclose(-val_pair[0], -val_pair[1])

    def test_iadd(self, val_pair, other_pair):
        sum_ = val_pair[0]
        sum_ += other_pair[0]
        assert np.isclose(get_val(sum_), val_pair[1] + other_pair[1])
        sum_ = other_pair[0]
        sum_ += val_pair[1]
        assert np.isclose(get_val(sum_), val_pair[1] + other_pair[1])

    def test_str(self, val_pair):
        assert isinstance(str(val_pair[0]), str)

    def test_repr(self, val_pair):
        assert isinstance(repr(val_pair[0]), str)

    def test_overflow(self):
        assert mici.utils.LogRepFloat(log_val=1e6) == np.inf

    def test_underflow(self):
        assert mici.utils.LogRepFloat(log_val=-1e6) == 0.0

    def test_neg_init(self):
        with pytest.raises(ValueError, match="non-negative"):
            mici.utils.LogRepFloat(val=-1)

    def test_no_init(self):
        with pytest.raises(ValueError, match="One of val or log_val must be specified"):
            mici.utils.LogRepFloat()

    def test_both_init(self):
        with pytest.raises(ValueError, match="Specify only one of val and log_val."):
            mici.utils.LogRepFloat(val=1.0, log_val=0.0)



"""Methods for adaptively setting algorithmic parameters of transitions."""

from __future__ import annotations

from abc import ABC, abstractmethod
from math import exp, log
from typing import TYPE_CHECKING

import numpy as np

from mici.errors import AdaptationError, IntegratorError
from mici.matrices import DensePositiveDefiniteMatrix, PositiveDiagonalMatrix

if TYPE_CHECKING:
    from collections.abc import Collection, Iterable

    from numpy.random import Generator
    from numpy.typing import ArrayLike

    from mici.integrators import Integrator
    from mici.states import ChainState
    from mici.systems import System
    from mici.transitions import Transition
    from mici.types import (
        AdaptationStatisticFunction,
        AdapterState,
        ReducerFunction,
        TransitionStatistics,
    )


class Adapter(ABC):
    """Abstract adapter for implementing schemes to adapt transition parameters.

    Adaptation schemes are assumed to be based on updating a collection of adaptation
    variables (collectively termed the adapter state here) after each chain transition
    based on the sampled chain state and/or statistics of the transition such as an
    acceptance probability statistic. After completing a chain of one or more adaptive
    transitions, the final adapter state may be used to perform a final update to the
    transition parameters.
    """

    @abstractmethod
    def initialize(
        self,
        chain_state: ChainState,
        transition: Transition,
    ) -> AdapterState:
        """Initialize adapter state prior to starting adaptive transitions.

        Args:
            chain_state: Initial chain state adaptive transition will be started from.
                May be used to calculate initial adapter state but should not be mutated
                by method.
            transition: Markov transition being adapted. Attributes of the transition or
                child objects may be updated in-place by the method.

        Returns:
            Initial adapter state.
        """

    @abstractmethod
    def update(
        self,
        adapt_state: AdapterState,
        chain_state: ChainState,
        trans_stats: TransitionStatistics,
        transition: Transition,
    ) -> None:
        """Update adapter state after sampling from transition being adapted.

        Args:
            adapt_state: Current adapter state. Entries will be updated in-place by the
                method.
            chain_state: Current chain state following sampling from transition being
                adapted. May be used to calculate adapter state updates but should not
                be mutated by method.
            trans_stats: Dictionary of statistics associated with transition being
                adapted. May be used to calculate adapter state updates but should not
                be mutated by method.
            transition: Markov transition being adapted. Attributes of the transition or
                child objects may be updated in-place by the method.
        """

    @abstractmethod
    def finalize(
        self,
        adapt_states: AdapterState | Iterable[AdapterState],
        chain_states: ChainState | Iterable[ChainState],
        transition: Transition,
        rngs: Generator | Iterable[Generator],
    ) -> None:
        """Update transition parameters based on final adapter state or states.

        Optionally, if multiple adapter states are available, e.g. from a set of
        independent adaptive chains, then these adaptation information from all the
        chains may be combined to set the transition parameter(s).

        Args:
            adapt_states: Final adapter state or a list of per chain adapter states.
                Arrays / buffers associated with the adapter state entries may be
                recycled to reduce memory usage - if so  the corresponding entries will
                be removed from the adapter state dictionary / dictionaries.
            chain_states: Final state of chain (or states of chains) in current
                sampling stage. May be updated in-place if transition parameters altered
                by adapter require updating any state components.
            transition: Markov transition being dapted. Attributes of the transition or
                child objects will be updated in-place by the method.
            rngs: Random number generator for the chain or a list of per-chain random
                number generators. Used to resample any components of states needing to
                be updated due to adaptation if required.
        """

    @property
    @abstractmethod
    def is_fast(self) -> bool:
        """Whether the adapter is 'fast' or 'slow'.

        An adapter which requires only local information to adapt the transition
        parameters should be classified as fast while one which requires more global
        information and so more chain iterations should be classified as slow i.e.
        :code:`is_fast == False`.
        """


def arithmetic_mean_log_step_size_reducer(log_step_sizes: Collection[float]) -> float:
    """Compute arithmetic mean of step sizes from their logs.

    Args:
        log_step_sizes: Logarithms of per-chain estimated step sizes.

    Returns:
        Arithmetic mean of estimated step sizes.
    """
    return sum(exp(x) for x in log_step_sizes) / len(log_step_sizes)


def geometric_mean_log_step_size_reducer(log_step_sizes: Collection[float]) -> float:
    """Compute geometric mean of step sizes from their logs.

    Args:
        log_step_sizes: Logarithms of per-chain estimated step sizes.

    Returns:
        Geometric mean of estimated step sizes.
    """
    return exp(sum(x for x in log_step_sizes) / len(log_step_sizes))


def min_log_step_size_reducer(log_step_sizes: Collection[float]) -> float:
    """Compute minimum of step sizes from their logs.

    Args:
        log_step_sizes: Logarithms of per-chain estimated step sizes.

    Returns:
        Minimum of estimated step sizes.
    """
    return exp(min(log_step_sizes))


def default_adapt_stat_func(stats: TransitionStatistics) -> float:
    """Function to extract default statistic used for step-size adaptation.

    Args:
        stats: Dictionary of transition statistics.

    Returns:
        Acceptance statistic.
    """
    return stats["accept_stat"]


class DualAveragingStepSizeAdapter(Adapter):
    """Dual averaging integrator step size adapter.

    Implementation of the dual algorithm step size adaptation algorithm described in
    Hoffman and Gelman (2014), a modified version of the stochastic optimisation scheme
    of Nesterov (2009). By default the adaptation is performed to control the
    :code:`accept_stat` statistic of an integration transition to be close to a target
    value but the statistic adapted on can be altered by changing the
    :code:`adapt_stat_func`.

    References:
      1. Hoffman, M.D. and Gelman, A. (2014). The No-U-turn sampler: adaptively setting
         path lengths in Hamiltonian Monte Carlo. Journal of Machine Learning Research,
         15(1), pp.1593-1623.
      2. Nesterov, Y. (2009). Primal-dual subgradient methods for convex problems.
         Mathematical programming 120(1), pp.221-259.
    """

    is_fast = True

    def __init__(
        self,
        adapt_stat_target: float = 0.8,
        adapt_stat_func: AdaptationStatisticFunction | None = None,
        log_step_size_reg_target: float | None = None,
        log_step_size_reg_coefficient: float = 0.05,
        iter_decay_coeff: float = 0.75,
        iter_offset: int = 10,
        max_init_step_size_iters: int = 100,
        log_step_size_reducer: ReducerFunction | None = None,
    ) -> None:
        """
        Args:
            adapt_stat_target: Target value for the transition statistic
                being controlled during adaptation.
            adapt_stat_func: Function which given a dictionary of transition statistics
                outputs the value of the statistic to control during adaptation. By
                default this is set to a function which simply selects the
                :code:'accept_stat' value in the statistics dictionary.
            log_step_size_reg_target: Value to regularize the controlled output
                (logarithm of the integrator step size) towards. If :code:`None` set to
                :code:`log(10 * init_step_size)` where :code:`init_step_size` is the
                initial 'reasonable' step size found by a coarse search as recommended
                in Hoffman and Gelman (2014). This has the effect of giving the dual
                averaging algorithm a tendency towards testing step sizes larger than
                the initial value, with typically integrating with a larger step size
                having a lower computational cost.
            log_step_size_reg_coefficient: Coefficient controlling amount of
                regularisation of controlled output (logarithm of the integrator step
                size) towards :code:`log_step_size_reg_target`. Defaults to 0.05 as
                recommended in Hoffman and Gelman (2014).
            iter_decay_coeff: Coefficient controlling exponent of decay in schedule
                weighting stochastic updates to smoothed log step size estimate. Should
                be in the interval (0.5, 1] to ensure asymptotic convergence of
                adaptation. A value of 1 gives equal weight to the whole history of
                updates while setting to a smaller value increasingly highly weights
                recent updates, giving a tendency to 'forget' early updates.
                Defaults to 0.75 as recommended in Hoffman and Gelman (2014).
            iter_offset: Offset used for the iteration based weighting of the adaptation
                statistic error estimate. Should be set to a non-negative value. A value
                > 0 has the effect of stabilising early iterations. Defaults to the
                value of 10 as recommended in Hoffman and Gelman (2014).
            max_init_step_size_iters: Maximum number of iterations to use in initial
                search for a reasonable step size with an
                :py:exc:`mici.errors.AdaptationError` exception raised if a suitable
                step size is not found within this many iterations.
            log_step_size_reducer: Reduction to apply to final per-chain step sizes
                estimates to produce overall integrator step size for main chain stages.
                The specified function should accept a sequence of logarithms of
                estimated step sizes and output a non-negative step size to use. If
                :code:`None`, the default, a function which computes the arithmetic mean
                of the per-chain step sizes is used.
        """
        self.adapt_stat_target = adapt_stat_target
        self.adapt_stat_func = (
            default_adapt_stat_func if adapt_stat_func is None else adapt_stat_func
        )
        self.log_step_size_reg_target = log_step_size_reg_target
        self.log_step_size_reg_coefficient = log_step_size_reg_coefficient
        self.iter_decay_coeff = iter_decay_coeff
        self.iter_offset = iter_offset
        self.max_init_step_size_iters = max_init_step_size_iters
        self.log_step_size_reducer = (
            arithmetic_mean_log_step_size_reducer
            if log_step_size_reducer is None
            else log_step_size_reducer
        )

    def initialize(
        self,
        chain_state: ChainState,
        transition: Transition,
    ) -> AdapterState:
        integrator = transition.integrator
        system = transition.system
        adapt_state = {
            "iter": 0,
            "smoothed_log_step_size": 0.0,
            "adapt_stat_error": 0.0,
        }
        init_step_size = self._find_and_set_init_step_size(
            chain_state,
            system,
            integrator,
        )
        if self.log_step_size_reg_target is None:
            adapt_state["log_step_size_reg_target"] = log(10 * init_step_size)
        else:
            adapt_state["log_step_size_reg_target"] = self.log_step_size_reg_target
        return adapt_state

    def _find_and_set_init_step_size(
        self,
        state: ChainState,
        system: System,
        integrator: Integrator,
    ) -> float:
        """Find initial step size by coarse search using single step statistics.

        Adaptation of Algorithm 4 in Hoffman and Gelman (2014).

        Compared to the Hoffman and Gelman algorithm, this version makes two changes:

          1. The absolute value of the change in Hamiltonian over a step being larger or
             smaller than log(2) is used to determine whether the step size is too big
             or small as opposed to the value of the equivalent Metropolis accept
             probability being larger or smaller than 0.5. Although a negative change in
             the Hamiltonian over a step of magnitude more than log(2) will lead to an
             accept probability of 1 for the forward move, the corresponding reversed
             move will have an accept probability less than 0.5, and so a change in the
             Hamiltonian over a step of magnitude more than log(2) irrespective of the
             sign of the change is indicative of the minimum acceptance probability over
             both forward and reversed steps being less than 0.5.
          2. To allow for integrators for which an integrator step may fail due to e.g.
             a convergence error in an iterative solver, the step size is also
             considered to be too big if any of the step sizes tried in the search
             result in a failed integrator step, with in this case the step size always
             being decreased on subsequent steps irrespective of the initial Hamiltonian
             error, until a integrator step successfully completes and the absolute
             value of the change in Hamiltonian is below the threshold of log(2)
             (corresponding to a minimum acceptance probability over forward and
             reversed steps of 0.5).
        """
        init_state = state.copy()
        h_init = system.h(init_state)
        if np.isnan(h_init):
            msg = "Hamiltonian evaluating to NaN at initial state."
            raise AdaptationError(msg)
        integrator.step_size = 1
        delta_h_threshold = log(2)
        for s in range(self.max_init_step_size_iters):
            try:
                state = integrator.step(init_state)
                delta_h = abs(h_init - system.h(state))
                if s == 0 or np.isnan(delta_h):
                    step_size_too_big = np.isnan(delta_h) or delta_h > delta_h_threshold
                if (step_size_too_big and delta_h <= delta_h_threshold) or (
                    not step_size_too_big and delta_h > delta_h_threshold
                ):
                    return integrator.step_size
                if step_size_too_big:
                    integrator.step_size /= 2
                else:
                    integrator.step_size *= 2
            except IntegratorError:  # noqa: PERF203
                step_size_too_big = True
                integrator.step_size /= 2
        msg = (
            f"Could not find reasonable initial step size in "
            f"{self.max_init_step_size_iters} iterations (final step size "
            f"{integrator.step_size}). A very large final step size may indicate that "
            f"the target distribution is improper such that the negative log density "
            f"is flat in one or more directions while a very small final step size may "
            f"indicate that the density function is insufficiently smooth at the point "
            f"initialized at."
        )
        raise AdaptationError(msg)

    def update(
        self,
        adapt_state: AdapterState,
        chain_state: ChainState,  # noqa: ARG002
        trans_stats: TransitionStatistics,
        transition: Transition,
    ) -> None:
        adapt_state["iter"] += 1
        error_weight = 1 / (self.iter_offset + adapt_state["iter"])
        adapt_state["adapt_stat_error"] *= 1 - error_weight
        adapt_state["adapt_stat_error"] += error_weight * (
            self.adapt_stat_target - self.adapt_stat_func(trans_stats)
        )
        smoothing_weight = (1 / adapt_state["iter"]) ** self.iter_decay_coeff
        log_step_size = adapt_state["log_step_size_reg_target"] - (
            adapt_state["adapt_stat_error"]
            * adapt_state["iter"] ** 0.5
            / self.log_step_size_reg_coefficient
        )
        adapt_state["smoothed_log_step_size"] *= 1 - smoothing_weight
        adapt_state["smoothed_log_step_size"] += smoothing_weight * log_step_size
        transition.integrator.step_size = exp(log_step_size)

    def finalize(
        self,
        adapt_states: AdapterState | Iterable[AdapterState],
        chain_states: ChainState | Iterable[ChainState],  # noqa: ARG002
        transition: Transition,
        rngs: Generator | Iterable[Generator],  # noqa: ARG002
    ) -> None:
        if isinstance(adapt_states, dict):
            transition.integrator.step_size = exp(
                adapt_states["smoothed_log_step_size"],
            )
        else:
            transition.integrator.step_size = self.log_step_size_reducer(
                [adapt_state["smoothed_log_step_size"] for adapt_state in adapt_states],
            )


class OnlineVarianceMetricAdapter(Adapter):
    """Diagonal metric adapter using online variance estimates.

    Uses Welford's algorithm (Welford, 1962) to stably compute an online estimate of the
    sample variances of the chain state position components during sampling. If online
    estimates are available from multiple independent chains, the final variance
    estimate is calculated from the per-chain statistics using the parallel / batched
    incremental variance algorithm described by Chan et al. (1979). The variance
    estimates are optionally regularized towards a common scalar value, with increasing
    weight for small number of samples, to decrease the effect of noisy estimates for
    small sample sizes, following the approach in Stan (Carpenter et al., 2017). The
    metric matrix representation is set to a diagonal matrix with diagonal elements
    corresponding to the reciprocal of the (regularized) variance estimates.

    References:
      1. Welford, B. P. (1962). Note on a method for calculating corrected sums of
         squares and products. Technometrics, 4(3), pp. 419-420.
      2. Chan, T. F., Golub, G. H. and LeVeque, R. J. (1979). Updating formulae and a
         pairwise algorithm for computing sample variances. Technical Report
         STAN-CS-79-773, Department of Computer Science, Stanford University.
      3. Carpenter, B., Gelman, A., Hoffman, M.D., Lee, D., Goodrich, B., Betancourt,
         M., Brubaker, M., Guo, J., Li, P. and Riddell, A.  (2017). Stan: A
         probabilistic programming language. Journal of Statistical Software, 76(1).
    """

    is_fast = False

    def __init__(self, reg_iter_offset: int = 5, reg_scale: float = 1e-3) -> None:
        """
        Args:
            reg_iter_offset: Iteration offset used for calculating iteration dependent
                weighting between regularisation target and current covariance estimate.
                Higher values cause stronger regularisation during initial iterations. A
                value of zero corresponds to no regularisation; this should only be used
                if the sample covariance is guaranteed to be positive definite.
            reg_scale: Positive scalar defining value variance estimates are regularized
                towards.
        """
        self.reg_iter_offset = reg_iter_offset
        self.reg_scale = reg_scale

    def initialize(
        self,
        chain_state: ChainState,
        transition: Transition,  # noqa: ARG002
    ) -> AdapterState:
        return {
            "iter": 0,
            "mean": np.zeros_like(chain_state.pos),
            "sum_diff_sq": np.zeros_like(chain_state.pos),
        }

    def update(
        self,
        adapt_state: AdapterState,
        chain_state: ChainState,
        trans_stats: TransitionStatistics,  # noqa: ARG002
        transition: Transition,  # noqa: ARG002
    ) -> None:
        # Use Welford (1962) incremental algorithm to update statistics to
        # calculate online variance estimate
        # https://en.wikipedia.org/wiki/
        #   Algorithms_for_calculating_variance#Welford's_online_algorithm
        adapt_state["iter"] += 1
        pos_minus_mean = chain_state.pos - adapt_state["mean"]
        adapt_state["mean"] += pos_minus_mean / adapt_state["iter"]
        adapt_state["sum_diff_sq"] += pos_minus_mean * (
            chain_state.pos - adapt_state["mean"]
        )

    def _regularize_var_est(self, var_est: ArrayLike, n_iter: int) -> None:
        """Update variance estimates by regularizing towards common scalar.

        Performed in place to prevent further array allocations.
        """
        if self.reg_iter_offset is not None and self.reg_iter_offset != 0:
            var_est *= n_iter / (self.reg_iter_offset + n_iter)
            var_est += self.reg_scale * (
                self.reg_iter_offset / (self.reg_iter_offset + n_iter)
            )

    def finalize(
        self,
        adapt_states: AdapterState | Iterable[AdapterState],
        chain_states: ChainState | Iterable[ChainState],
        transition: Transition,
        rngs: Generator | Iterable[Generator],
    ) -> None:
        if isinstance(adapt_states, dict):
            n_iter = adapt_states["iter"]
            var_est = adapt_states.pop("sum_diff_sq")
            chain_states = [chain_states]
            rngs = [rngs]
        else:
            # Use Chan et al. (1979) parallel variance estimation algorithm
            # to combine per-chain statistics
            # https://en.wikipedia.org/wiki/
            #    Algorithms_for_calculating_variance#Parallel_algorithm
            for i, adapt_state in enumerate(adapt_states):
                if i == 0:
                    n_iter = adapt_state["iter"]
                    mean_est = adapt_state.pop("mean")
                    var_est = adapt_state.pop("sum_diff_sq")
                else:
                    n_iter_prev = n_iter
                    n_iter += adapt_state["iter"]
                    mean_diff = mean_est - adapt_state["mean"]
                    mean_est *= n_iter_prev
                    mean_est += adapt_state["iter"] * adapt_state["mean"]
                    mean_est /= n_iter
                    var_est += adapt_state["sum_diff_sq"]
                    var_est += (
                        mean_diff**2 * (adapt_state["iter"] * n_iter_prev) / n_iter
                    )
        if n_iter < 2:  # noqa: PLR2004
            msg = "At least two chain samples required to compute a variance estimates."
            raise AdaptationError(msg)
        var_est /= n_iter - 1
        self._regularize_var_est(var_est, n_iter)
        transition.system.metric = PositiveDiagonalMatrix(var_est).inv
        # Resample momentum to account for altered distribution due to new metric
        for chain_state, rng in zip(chain_states, rngs, strict=True):
            chain_state.mom = transition.system.sample_momentum(chain_state, rng)


class OnlineCovarianceMetricAdapter(Adapter):
    """Dense metric adapter using online covariance estimates.

    Uses Welford's algorithm (Welford, 1962) to stably compute an online estimate of the
    sample covariance matrix of the chain state position components during sampling. If
    online estimates are available from multiple independent chains, the final
    covariance matrix estimate is calculated from the per-chain statistics using a
    covariance variant due to Schubert and Gertz (2018) of the parallel / batched
    incremental variance algorithm described by Chan et al. (1979). The covariance
    matrix estimates are optionally regularized towards a scaled identity matrix, with
    increasing weight for small number of samples, to decrease the effect of noisy
    estimates for small sample sizes, following the approach in Stan (Carpenter et al.,
    2017). The metric matrix representation is set to a dense positive definite matrix
    corresponding to the inverse of the (regularized) covariance matrix estimate.


    References:
      1. Welford, B. P. (1962). Note on a method for calculating corrected sums of
         squares and products. Technometrics, 4(3), pp. 419-420.
      2. Schubert, E. and Gertz, M. (2018). Numerically stable parallel computation of
         (co-)variance. ACM. p. 10. doi:10.1145/3221269.3223036.
      3. Chan, T. F., Golub, G. H. and LeVeque, R. J. (1979). Updating formulae and a
         pairwise algorithm for computing sample variances. Technical Report
         STAN-CS-79-773, Department of Computer Science, Stanford University.
      4. Carpenter, B., Gelman, A., Hoffman, M.D., Lee, D., Goodrich, B., Betancourt,
         M., Brubaker, M., Guo, J., Li, P. and Riddell, A.  (2017). Stan: A
         probabilistic programming language. Journal of Statistical Software, 76(1).
    """

    is_fast = False

    def __init__(self, reg_iter_offset: int = 5, reg_scale: float = 1e-3) -> None:
        """
        Args:
            reg_iter_offset: Iteration offset used for calculating iteration
                dependent weighting between regularisation target and current covariance
                estimate. Higher values cause stronger regularisation during initial
                iterations.
            reg_scale: Positive scalar defining value variance estimates are
                regularized towards.
        """
        self.reg_iter_offset = reg_iter_offset
        self.reg_scale = reg_scale

    def initialize(
        self,
        chain_state: ChainState,
        transition: Transition,  # noqa: ARG002
    ) -> AdapterState:
        dim_pos = chain_state.pos.shape[0]
        dtype = chain_state.pos.dtype
        return {
            "iter": 0,
            "mean": np.zeros(shape=(dim_pos,), dtype=dtype),
            "sum_diff_outer": np.zeros(shape=(dim_pos, dim_pos), dtype=dtype),
        }

    def update(
        self,
        adapt_state: AdapterState,
        chain_state: ChainState,
        trans_stats: TransitionStatistics,  # noqa: ARG002
        transition: Transition,  # noqa: ARG002
    ) -> None:
        # Use Welford (1962) incremental algorithm to update statistics to
        # calculate online covariance estimate
        # https://en.wikipedia.org/wiki/
        #  Algorithms_for_calculating_variance#Online
        adapt_state["iter"] += 1
        pos_minus_mean = chain_state.pos - adapt_state["mean"]
        adapt_state["mean"] += pos_minus_mean / adapt_state["iter"]
        adapt_state["sum_diff_outer"] += (
            pos_minus_mean[None, :] * (chain_state.pos - adapt_state["mean"])[:, None]
        )

    def _regularize_covar_est(self, covar_est: ArrayLike, n_iter: int) -> None:
        """Update covariance estimate by regularising towards identity.

        Performed in place to prevent further array allocations.
        """
        covar_est *= n_iter / (self.reg_iter_offset + n_iter)
        covar_est_diagonal = np.einsum("ii->i", covar_est)
        covar_est_diagonal += self.reg_scale * (
            self.reg_iter_offset / (self.reg_iter_offset + n_iter)
        )

    def finalize(
        self,
        adapt_states: AdapterState | Iterable[AdapterState],
        chain_states: ChainState | Iterable[ChainState],
        transition: Transition,
        rngs: Generator | Iterable[Generator],
    ) -> None:
        if isinstance(adapt_states, dict):
            n_iter = adapt_states["iter"]
            covar_est = adapt_states.pop("sum_diff_outer")
            chain_states = [chain_states]
            rngs = [rngs]
        else:
            # Use Schubert and Gertz (2018) parallel covariance estimation
            # algorithm to combine per-chain statistics
            for i, adapt_state in enumerate(adapt_states):
                if i == 0:
                    n_iter = adapt_state["iter"]
                    mean_est = adapt_state.pop("mean")
                    covar_est = adapt_state.pop("sum_diff_outer")
                else:
                    n_iter_prev = n_iter
                    n_iter += adapt_state["iter"]
                    mean_diff = mean_est - adapt_state["mean"]
                    mean_est *= n_iter_prev
                    mean_est += adapt_state["iter"] * adapt_state["mean"]
                    mean_est /= n_iter
                    covar_est += adapt_state["sum_diff_outer"]
                    covar_est += (
                        np.outer(mean_diff, mean_diff)
                        * (adapt_state["iter"] * n_iter_prev)
                        / n_iter
                    )
        if n_iter < 2:  # noqa: PLR2004
            msg = "At least two chain samples required to compute a variance estimates."
            raise AdaptationError(msg)
        covar_est /= n_iter - 1
        self._regularize_covar_est(covar_est, n_iter)
        transition.system.metric = DensePositiveDefiniteMatrix(covar_est).inv
        # Resample momentum to account for altered distribution due to new metric
        for chain_state, rng in zip(chain_states, rngs, strict=True):
            chain_state.mom = transition.system.sample_momentum(chain_state, rng)



"""Progress bar classes for tracking progress of chains."""

from __future__ import annotations

import abc
import html
import importlib
import sys
from timeit import default_timer as timer
from typing import TYPE_CHECKING, Protocol

try:
    from IPython import get_ipython
    from IPython.display import display as ipython_display

    IPYTHON_AVAILABLE = True
except ImportError:
    IPYTHON_AVAILABLE = False

if TYPE_CHECKING:
    from collections.abc import Collection, Generator
    from queue import Queue
    from types import TracebackType
    from typing import Any, TextIO

    from typing_extensions import Self


ON_COLAB = (
    importlib.util.find_spec("google") is not None
    and importlib.util.find_spec("google.colab") is not None
)


def _in_interactive_shell() -> bool:
    """Check if in interactive shell which supports updateable displays."""
    if not IPYTHON_AVAILABLE:
        return False
    if ON_COLAB:
        return True
    try:
        ipython = get_ipython()
        ipython_module = ipython.__module__
        ipython_class = ipython.__class__.__name__
    except NameError:
        return False
    return (ipython_module, ipython_class) in {
        ("ipykernel.zmqshell", "ZMQInteractiveShell"),
        ("pyodide_kernel.interpreter", "Interpreter"),
    }


class UpdateableDisplay(Protocol):
    """Updateable display."""

    def update(self, obj: Any) -> None:  # noqa: ANN401
        """Update display with representation of object."""


def _create_display(
    obj: Any,  # noqa: ANN401
    position: tuple[int, int],
) -> UpdateableDisplay:
    """Create an updateable display object.

    Args:
        obj: Initial object to display.
        position: Tuple specifying position of display within a sequence of displays
            with first entry corresponding to the zero-indexed position and the second
            entry the total number of displays.

    Returns:
        Object with `update` method to update displayed content.
    """
    if _in_interactive_shell():
        return ipython_display(obj, display_id=True)
    return FileDisplay(position)


def _format_time(total_seconds: float) -> str:
    """Format a time interval in seconds as a colon-delimited string [h:]m:s."""
    total_mins, seconds = divmod(int(total_seconds), 60)
    hours, mins = divmod(total_mins, 60)
    if hours != 0:
        return f"{hours:d}:{mins:02d}:{seconds:02d}"
    return f"{mins:02d}:{seconds:02d}"


def _update_stats_running_means(
    iter_: int,
    means: dict[str, float],
    new_vals: dict[str, float],
) -> None:
    """Update dictionary of running statistics means with latest values."""
    if iter_ == 1:
        means.update({key: float(val) for key, val in new_vals.items()})
    else:
        for key, val in new_vals.items():
            means[key] += (float(val) - means[key]) / iter_


class ProgressBar(abc.ABC):
    """Base class defining expected interface for progress bars."""

    def __init__(
        self,
        sequence: Collection,
        description: str | None,
        position: tuple[int, int] = (0, 1),
    ) -> None:
        """
        Args:
            sequence: Sequence to iterate over. Must be iterable _and_ have a defined
                length such that `len(sequence)` is valid.
            description: Description of task to prefix progress bar with.
            position: Tuple specifying position of progress bar within a sequence with
                first entry corresponding to zero-indexed position and the second entry
                the total number of progress bars.
        """
        self._sequence = sequence
        self._description = description
        self._position = position
        self._active = False
        self._n_iter = len(sequence)

    @property
    def sequence(self) -> Collection:
        """Sequence iterated over."""
        return self._sequence

    @sequence.setter
    def sequence(self, value: Collection) -> None:
        if self._active:
            msg = "Cannot set sequence of active progress bar."
            raise RuntimeError(msg)
        self._sequence = value
        self._n_iter = len(value)

    @property
    def n_iter(self) -> int:
        return self._n_iter

    def __iter__(self) -> Generator[tuple[Any, dict[str, float]], None, None]:
        for i, val in enumerate(self.sequence):
            iter_dict = {}
            yield val, iter_dict
            self.update(i + 1, iter_dict, refresh=True)

    def __len__(self) -> int:
        return self._n_iter

    @abc.abstractmethod
    def update(
        self,
        iter_count: int,
        iter_dict: dict[str, float] | None,
        *,
        refresh: bool = True,
    ) -> None:
        """Update progress bar state.

        Args:
            iter_count: New value for iteration counter.
            iter_dict: Dictionary of iteration statistics key-value pairs to use to
                update postfix stats.
            refresh: Whether to refresh display(s).
        """

    def __enter__(self) -> Self:
        """Set up progress bar and any associated resource."""
        self._active = True
        return self

    def __exit__(
        self,
        _exc_type: None | type[BaseException],
        _exc_value: None | BaseException,
        _traceback: None | TracebackType,
    ) -> bool:
        """Close down progress bar and any associated resources."""
        self._active = False
        return False


class DummyProgressBar(ProgressBar):
    """Placeholder progress bar which does not display progress updates."""

    def update(
        self,
        iter_count: int,
        iter_dict: dict[str, float] | None,
        *,
        refresh: bool = True,
    ) -> None:
        pass


class SequenceProgressBar(ProgressBar):
    """Iterable object for tracking progress of an iterative task.

    Implements both string and HTML representations to allow richer
    display in interfaces which support HTML output, for example Jupyter
    notebooks or interactive terminals.
    """

    GLYPHS = " ▏▎▍▌▋▊▉█"
    """Characters used to create string representation of progress bar."""

    def __init__(
        self,
        sequence: Collection,
        description: str | None = None,
        position: tuple[int, int] = (0, 1),
        displays: Collection | None = None,
        n_col: int = 10,
        unit: str = "it",
        min_refresh_time: float = 0.25,
    ) -> None:
        """
        Args:
            sequence: Sequence to iterate over. Must be iterable **and** have a defined
                length such that `len(sequence)` is valid.
            description: Description of task to prefix progress bar with.
            position: Tuple specifying position of progress bar within a sequence with
                first entry corresponding to zero-indexed position and the second entry
                the total number of progress bars.
            displays: List of objects to use to display visual representation(s) of
                progress bar. Each object much have an `update` method which will be
                passed a single argument corresponding to the current progress bar.
            n_col: Number of columns (characters) to use in string representation of
                progress bar.
            unit: String describing unit of per-iteration tasks.
            min_referesh_time: Minimum time in seconds between each refresh of progress
                bar visual representation.
        """
        super().__init__(sequence, description, position)
        self._n_col = n_col
        self._unit = unit
        self._counter = 0
        self._start_time = None
        self._elapsed_time = 0
        self._stats_dict = {}
        self._displays = displays
        self._min_refresh_time = min_refresh_time

    @property
    def description(self) -> str:
        """Description of task being tracked."""
        return self._description

    @property
    def counter(self) -> int:
        """Progress iteration count."""
        return self._counter

    @counter.setter
    def counter(self, value: int) -> None:
        self._counter = max(0, min(value, self.n_iter))

    @property
    def prop_complete(self) -> float:
        """Proportion complete (float value in [0, 1])."""
        return self.counter / self.n_iter

    @property
    def perc_complete(self) -> str:
        """Percentage complete formatted as string."""
        return f"{int(self.prop_complete * 100):3d}%"

    @property
    def elapsed_time(self) -> str:
        """Elapsed time formatted as string."""
        return _format_time(self._elapsed_time)

    @property
    def iter_rate(self) -> str:
        """Mean iteration rate if ≥ 1 `it/s` or reciprocal `s/it` as string."""
        if self.prop_complete == 0:
            return "?"
        mean_time = self._elapsed_time / self.counter
        return (
            f"{mean_time:.2f}s/{self._unit}"
            if mean_time > 1
            else f"{1/mean_time:.2f}{self._unit}/s"
        )

    @property
    def est_remaining_time(self) -> str:
        """Estimated remaining time to completion formatted as string."""
        if self.prop_complete == 0:
            return "?"
        return _format_time((1 / self.prop_complete - 1) * self._elapsed_time)

    @property
    def n_block_filled(self) -> int:
        """Number of filled blocks in progress bar."""
        return int(self._n_col * self.prop_complete)

    @property
    def n_block_empty(self) -> int:
        """Number of empty blocks in progress bar."""
        return self._n_col - self.n_block_filled

    @property
    def prop_partial_block(self) -> float:
        """Proportion filled in partial block in progress bar."""
        return self._n_col * self.prop_complete - self.n_block_filled

    @property
    def filled_blocks(self) -> str:
        """Filled blocks string."""
        return self.GLYPHS[-1] * self.n_block_filled

    @property
    def empty_blocks(self) -> str:
        """Empty blocks string."""
        if self.prop_partial_block == 0:
            return self.GLYPHS[0] * self.n_block_empty
        return self.GLYPHS[0] * (self.n_block_empty - 1)

    @property
    def partial_block(self) -> str:
        """Partial block character."""
        if self.prop_partial_block == 0:
            return ""
        return self.GLYPHS[int(len(self.GLYPHS) * self.prop_partial_block)]

    @property
    def progress_bar(self) -> str:
        """Progress bar string."""
        return f"|{self.filled_blocks}{self.partial_block}{self.empty_blocks}|"

    @property
    def bar_color(self) -> str:
        """CSS color property for HTML progress bar."""
        if self.counter == self.n_iter:
            return "var(--jp-success-color1, #4caf50)"
        if self._active:
            return "var(--jp-brand-color1, #2196f3)"
        return "var(--jp-error-color1, #f44336)"

    @property
    def stats(self) -> str:
        """Comma-delimited string list of statistic key=value pairs."""
        return ", ".join(f"{k}={v:#.3g}" for k, v in self._stats_dict.items())

    @property
    def prefix(self) -> str:
        """Text to prefix progress bar with."""
        return (
            f'{self.description + ": "if self.description else ""}'
            f"{self.perc_complete}"
        )

    @property
    def postfix(self) -> str:
        """Text to postfix progress bar with."""
        return (
            f"{self.counter}/{self.n_iter} "
            f"[{self.elapsed_time}<{self.est_remaining_time}, "
            f"{self.iter_rate}"
            f'{", " + self.stats if self._stats_dict else ""}]'
        )

    def reset(self) -> None:
        """Reset progress bar state."""
        self._counter = 0
        self._start_time = timer()
        self._last_refresh_time = -float("inf")
        self._stats_dict = {}

    def update(
        self,
        iter_count: int,
        iter_dict: dict[str, float] | None = None,
        *,
        refresh: bool = True,
    ) -> None:
        """Update progress bar state.

        Args:
            iter_count: New value for iteration counter.
            iter_dict: Dictionary of iteration statistics key-value pairs to use to
                update postfix stats.
            refresh: Whether to refresh display(s).
        """
        if iter_count == 0:
            self.reset()
        else:
            self.counter = iter_count
            if iter_dict is not None:
                _update_stats_running_means(iter_count, self._stats_dict, iter_dict)
            self._elapsed_time = timer() - self._start_time
        if (
            refresh
            and iter_count == self.n_iter
            or (timer() - self._last_refresh_time > self._min_refresh_time)
        ):
            self.refresh()
            self._last_refresh_time = timer()

    def refresh(self) -> None:
        """Refresh visual display(s) of progress bar."""
        for display in self._displays:
            display.update(self)

    def __str__(self) -> str:
        return f"{self.prefix}{self.progress_bar}{self.postfix}"

    def __repr__(self) -> str:
        return self.__str__()

    def _repr_html_(self) -> str:
        return f"""
        <div style="line-height: 28px; width: 100%; display: flex;
                    flex-flow: row wrap; align-items: center;
                    position: relative; margin: 0px;">
          <label style="margin-right: 8px; flex-shrink: 0;
                        font-size: var(--jp-code-font-size, 13px);
                        font-family: var(--jp-code-font-family, monospace);">
            {html.escape(self.prefix).replace(' ', '&nbsp;')}
          </label>
          <div role="progressbar" aria-valuenow="{self.prop_complete}"
               aria-valuemin="0" aria-valuemax="1"
               style="position: relative; flex-grow: 1; align-self: stretch;
                      margin-top: 4px; margin-bottom: 4px;  height: initial;
                      background-color: #eee;">
            <div style="background-color: {self.bar_color}; position: absolute;
                        bottom: 0; left: 0; width: {self.perc_complete};
                        height: 100%;"></div>
          </div>
          <div style="margin-left: 8px; flex-shrink: 0;
                      font-family: var(--jp-code-font-family, monospace);
                      font-size: var(--jp-code-font-size, 13px);">
            {html.escape(self.postfix)}
          </div>
        </div>
        """

    def __enter__(self) -> Self:
        super().__enter__()
        self.reset()
        if self._displays is None:
            self._displays = [_create_display(self, self._position)]
        return self

    def __exit__(
        self,
        exc_type: None | type[BaseException],
        exc_value: None | BaseException,
        traceback: None | TracebackType,
    ) -> bool:
        ret_val = super().__exit__(exc_type, exc_value, traceback)
        if self.counter != self.n_iter:
            self.refresh()
        return ret_val


class LabelledSequenceProgressBar(ProgressBar):
    """Iterable object for tracking progress of a sequence of labelled tasks."""

    def __init__(
        self,
        labelled_sequence: dict[str, Any],
        description: str | None = None,
        position: tuple[int, int] = (0, 1),
        displays: Collection | None = None,
    ) -> None:
        """
        Args:
            labelled_sequence: Ordered dictionary with string keys corresponding to
                labels for stages represented by sequence and values the entries in the
                sequence being iterated over.
            description: Description of task to prefix progress bar with.
            position: Tuple specifying position of progress bar within a sequence with
                first entry corresponding to zero-indexed position and the second entry
                the total number of progress bars.
            displays: List of objects to use to display visual representation(s) of
                progress bar. Each object much have an `update` method which will be
                passed a single argument corresponding to the current progress bar.
        """
        super().__init__(list(labelled_sequence.values()), description, position)
        self._labels = list(labelled_sequence.keys())
        self._description = description
        self._position = position
        self._counter = 0
        self._prev_time = None
        self._iter_times = [None] * self.n_iter
        self._stats_dict = {}
        self._displays = displays

    @property
    def counter(self) -> int:
        """Progress iteration count."""
        return self._counter

    @counter.setter
    def counter(self, value: int) -> None:
        self._counter = max(0, min(value, self.n_iter))

    @property
    def description(self) -> str:
        """Description of task being tracked."""
        return self._description

    @property
    def stats(self) -> str:
        """Comma-delimited string list of statistic key=value pairs."""
        return ", ".join(f"{k}={v:#.3g}" for k, v in self._stats_dict.items())

    @property
    def prefix(self) -> str:
        """Text to prefix progress bar with."""
        return f'{self.description + ": " if self.description else ""}'

    @property
    def postfix(self) -> str:
        """Text to postfix progress bar with."""
        return f" [{self.stats}]" if self._stats_dict else ""

    @property
    def completed_labels(self) -> list[str]:
        """Labels corresponding to completed iterations."""
        return [
            f"{label} [{_format_time(time)}]"
            for label, time in zip(
                self._labels[: self._counter],
                self._iter_times[: self._counter],
                strict=True,
            )
        ]

    @property
    def current_label(self) -> str:
        """Label corresponding to current iteration."""
        return self._labels[self._counter] if self.counter < self.n_iter else ""

    @property
    def unstarted_labels(self) -> list[str]:
        """Labels corresponding to unstarted iterations."""
        return self._labels[self._counter + 1 :]

    @property
    def progress_bar(self) -> str:
        """Progress bar string."""
        labels = self.completed_labels
        if self.counter < self.n_iter:
            labels.append(self.current_label)
        return " > ".join(labels)

    def reset(self) -> None:
        """Reset progress bar state."""
        self._counter = 0
        self._prev_time = timer()
        self._iter_times = [None] * self.n_iter
        self._stats_dict = {}

    def update(
        self,
        iter_count: int,
        iter_dict: dict[str, float] | None = None,
        *,
        refresh: bool = True,
    ) -> None:
        """Update progress bar state.

        Args:
            iter_count: New value for iteration counter.
            iter_dict: Dictionary of iteration statistics key-value pairs to use to
                update postfix stats.
            refresh: Whether to refresh display(s).
        """
        if iter_count == 0:
            self.reset()
        else:
            self.counter = iter_count
            if iter_dict is not None:
                _update_stats_running_means(iter, self._stats_dict, iter_dict)
            curr_time = timer()
            self._iter_times[iter_count - 1] = curr_time - self._prev_time
            self._prev_time = curr_time
        if refresh:
            self.refresh()

    def refresh(self) -> None:
        """Refresh visual display(s) of status bar."""
        for display in self._displays:
            display.update(self)

    def __str__(self) -> str:
        return f"{self.prefix}{self.progress_bar}{self.postfix}"

    def __repr__(self) -> str:
        return self.__str__()

    def _repr_html_(self) -> str:
        html_string = f"""
        <div style="line-height: 24px; width: 100%; display: flex;
                    flex-flow: row wrap; align-items: center;
                    position: relative; margin: 0px;">
          <label style="flex-shrink: 0;
                        font-size: var(--jp-code-font-size, 13px);
                        font-family: var(--jp-code-font-family, monospace);">
            {html.escape(self.prefix).replace(' ', '&nbsp;')}
          </label>
        """
        template_string = """
          <div style="position: relative; flex-grow: 1; align-self: stretch;
                      margin: 1px; padding: 0px; text-align: center;
                      height: initial; background-color: {background_color};
                      color: {foreground_color}; border-radius: 5px;
                      border: 1px solid {foreground_color}; font-size: 90%;">
            {label}
          </div>
        """
        for label in self.completed_labels:
            html_string += template_string.format(
                label=label,
                foreground_color="white",
                background_color="#4caf50",
            )
        if self.counter != self.n_iter:
            html_string += template_string.format(
                label=self.current_label,
                foreground_color="white",
                background_color="#2196f3" if self._active else "#f44336",
            )
        for label in self.unstarted_labels:
            html_string += template_string.format(
                label=label,
                foreground_color="#aaa",
                background_color="white",
            )
        if self.postfix != "":
            html_string += f"""
              <div style="margin-left: 8px; flex-shrink: 0;
                          font-family: var(--jp-code-font-family, monospace);
                          font-size: var(--jp-code-font-size, 13px);">
                {html.escape(self.postfix)}
              </div>
            """
        html_string += "</div>"
        return html_string

    def __enter__(self) -> Self:
        super().__enter__()
        self.reset()
        if self._displays is None:
            self._displays = [_create_display(self, self._position)]
        self.refresh()
        return self

    def __exit__(
        self,
        exc_type: None | type[BaseException],
        exc_value: None | BaseException,
        traceback: None | TracebackType,
    ) -> bool:
        ret_val = super().__exit__(exc_type, exc_value, traceback)
        if self.counter != self.n_iter:
            self.refresh()
        return ret_val


class FileDisplay:
    """Use file which supports ANSI escape sequences as an updatable display."""

    CURSOR_UP = "\x1b[A"
    """ANSI escape sequence to move cursor up one line."""

    CURSOR_DOWN = "\x1b[B"
    """ANSI escape sequence to move cursor down one line."""

    def __init__(
        self,
        position: tuple[int, int] = (0, 1),
        file: TextIO | None = None,
    ) -> None:
        r"""
        Args:
            position: Tuple specifying position of display line within a sequence lines
                with first entry corresponding to zero-indexed line and the second entry
                the total number of lines.
            file: File object to write updates to. Must support ANSI escape sequences
                `\x1b[A}` (cursor up) and `\\x1b[B` (cursor down) for manipulating
                write position. Defaults to `sys.stdout` if `None`.
        """
        self._position = position
        self._file = file if file is not None else sys.stdout
        self._last_string_length = 0
        if self._position[0] == 0:
            self._file.write("\n" * self._position[1])
        self._file.flush()

    def _move_line(self, offset: int) -> None:
        self._file.write(self.CURSOR_DOWN * offset + self.CURSOR_UP * -offset)
        self._file.flush()

    def update(self, obj: Any) -> None:  # noqa: ANN401
        """Update display with string representation of object.

        Args:
            obj: Object to display.
        """
        self._move_line(self._position[0] - self._position[1])
        string = str(obj)
        self._file.write(f"{string: <{self._last_string_length}}\r")
        self._last_string_length = len(string)
        self._move_line(self._position[1] - self._position[0])
        self._file.flush()


class _ProxySequenceProgressBar:
    """Proxy progress bar that outputs progress updates to a queue.

    Intended for communicating progress updates from a child to parent process
    when distributing tasks across multiple processes.
    """

    def __init__(self, sequence: Collection, job_id: int, iter_queue: Queue) -> None:
        """
        Args:
            sequence: Sequence to iterate over. Must be iterable _and_ have a defined
                length such that `len(sequence)` is valid.
            job_id: Unique integer identifier for progress bar amongst other progress
                bars sharing same `iter_queue` object.
            iter_queue: Shared queue object that progress updates are pushed to.
        """
        self._sequence = sequence
        self._n_iter = len(sequence)
        self._job_id = job_id
        self._iter_queue = iter_queue

    def __len__(self) -> int:
        return self._n_iter

    def __enter__(self) -> Self:
        self._iter_queue.put((self._job_id, 0, None))
        return self

    def __exit__(
        self,
        exc_type: None | type[BaseException],
        exc_value: None | BaseException,
        traceback: None | TracebackType,
    ) -> bool:
        return False

    def __iter__(self) -> Generator[tuple[Any, dict[str, float]], None, None]:
        for i, val in enumerate(self._sequence):
            iter_dict = {}
            yield val, iter_dict
            self._iter_queue.put((self._job_id, i + 1, iter_dict))



import importlib

import numpy as np
import pytest

import mici

try:
    import arviz

    ARVIZ_AVAILABLE = importlib.util.find_spec("arviz") is not None
except ImportError:
    ARVIZ_AVAILABLE = False

try:
    import pymc

    PYMC_AVAILABLE = True
except ImportError:
    PYMC_AVAILABLE = False


try:
    import stan  # noqa: F401

    STAN_AVAILABLE = True
except ImportError:
    STAN_AVAILABLE = False


SEED = 3046987125
NUM_CHAIN = 2
NUM_SAMPLE = 10


@pytest.fixture
def rng():
    return np.random.default_rng(SEED)


if ARVIZ_AVAILABLE:

    def test_convert_to_inference_data(rng):
        traces = {
            "a": list(rng.standard_normal((NUM_CHAIN, NUM_SAMPLE))),
            "b": list(rng.standard_normal((NUM_CHAIN, NUM_SAMPLE))),
            "energy": list(rng.standard_normal((NUM_CHAIN, NUM_SAMPLE))),
            "lp": list(rng.standard_normal((NUM_CHAIN, NUM_SAMPLE))),
        }
        stats = {
            "accept_stat": list(rng.uniform(0, 1, (NUM_CHAIN, NUM_SAMPLE))),
            "n_step": list(rng.integers(0, 10, (NUM_CHAIN, NUM_SAMPLE))),
        }
        inference_data = mici.interop.convert_to_inference_data(traces, stats)
        assert isinstance(inference_data, arviz.InferenceData)
        groups = inference_data.groups()
        assert "posterior" in groups
        assert "sample_stats" in groups
        for group in groups:
            assert inference_data[group].coords["chain"].size == NUM_CHAIN
            assert inference_data[group].coords["draw"].size == NUM_SAMPLE
        assert set(inference_data.posterior.keys()) == set(traces.keys())
        assert set(inference_data.sample_stats.keys()) == {
            "acceptance_rate",
            "n_steps",
            "energy",
            "lp",
        }


if PYMC_AVAILABLE:

    @pytest.fixture
    def pymc_model():
        with pymc.Model() as model:
            pymc.Normal("x", mu=0, sigma=1)
            pymc.Uniform("y", lower=0, upper=1)
        return model

    @pytest.mark.parametrize("init", ["adapt_diag", "jitter+adapt_diag", "adapt_full"])
    @pytest.mark.parametrize("progressbar", [True, False])
    def test_sample_pymc_model(pymc_model, init, progressbar):
        traces = mici.interop.sample_pymc_model(
            model=pymc_model,
            draws=NUM_SAMPLE,
            tune=NUM_SAMPLE,
            random_seed=SEED,
            init=init,
            progressbar=progressbar,
            cores=1,
        )
        assert isinstance(traces, dict)
        for var in pymc_model.unobserved_RVs:
            assert var.name in traces
            assert traces[var.name].shape[1] == NUM_SAMPLE
            assert not np.any(np.isnan(traces[var.name]))

    if ARVIZ_AVAILABLE:

        def test_sample_pymc_model_inference_data(pymc_model):
            inference_data = mici.interop.sample_pymc_model(
                model=pymc_model,
                chains=NUM_CHAIN,
                draws=NUM_SAMPLE,
                random_seed=SEED,
                return_inferencedata=True,
                cores=1,
            )
            assert isinstance(inference_data, arviz.InferenceData)
            for var in pymc_model.unobserved_RVs:
                assert var.name in inference_data.posterior
                assert not np.any(np.isnan(inference_data.posterior[var.name]))
            assert inference_data.posterior.coords["chain"].size == NUM_CHAIN
            assert inference_data.posterior.coords["draw"].size == NUM_SAMPLE


if STAN_AVAILABLE:

    @pytest.fixture
    def stan_model_code_data_and_params():
        model_code = """
        parameters {
            real x;
            real <lower=0, upper=1> y;
        }
        model {
            x ~ normal(0, 1);
            y ~ uniform(0, 1);
        }
        """
        return model_code, {}, {"x", "y"}

    @pytest.mark.parametrize("metric", ["unit_e", "diag_e", "dense_e"])
    @pytest.mark.parametrize("adapt_engaged", [True, False])
    def test_sample_stan_model(stan_model_code_data_and_params, metric, adapt_engaged):
        model_code, data, params = stan_model_code_data_and_params
        traces = mici.interop.sample_stan_model(
            model_code=model_code,
            data=data,
            num_chains=NUM_CHAIN,
            num_samples=NUM_SAMPLE,
            num_warmup=NUM_SAMPLE,
            seed=SEED,
            metric=metric,
            adapt_engaged=adapt_engaged,
        )
        assert isinstance(traces, dict)
        for param in params:
            assert param in traces
            assert traces[param].shape[0] == NUM_SAMPLE * NUM_CHAIN
            assert not np.any(np.isnan(traces[param]))

    if ARVIZ_AVAILABLE:

        def test_sample_stan_model_inference_data(stan_model_code_data_and_params):
            model_code, data, params = stan_model_code_data_and_params
            inference_data = mici.interop.sample_stan_model(
                model_code=model_code,
                data=data,
                num_chains=NUM_CHAIN,
                num_samples=NUM_SAMPLE,
                num_warmup=NUM_SAMPLE,
                seed=SEED,
                return_inferencedata=True,
            )
            assert isinstance(inference_data, arviz.InferenceData)
            for param in params:
                assert param in inference_data.posterior
                assert not np.any(np.isnan(inference_data.posterior[param]))
            assert inference_data.posterior.coords["chain"].size == NUM_CHAIN
            assert inference_data.posterior.coords["draw"].size == NUM_SAMPLE



"""Exception types."""

from __future__ import annotations


class Error(RuntimeError):
    """Base class for errors."""


class IntegratorError(Error):
    """Error raised when integrator step fails."""


class NonReversibleStepError(IntegratorError):
    """Error raised when integrator step fails reversibility check."""


class ConvergenceError(IntegratorError):
    """Error raised when solver fails to converge within allowed iterations."""


class LinAlgError(Error):
    """Error raised when a matrix operation raises a linear algebra error."""


class HamiltonianDivergenceError(IntegratorError):
    """Error raised when integration of Hamiltonian dynamics diverges."""


class AdaptationError(Error):
    """Error raised when adaptation of transition parameters fails."""


class ReadOnlyStateError(Error):
    """Error raised when writing to attributes of read-only chain state."""
